import k3d
import numbers
import collections
import matplotlib
import numpy as np
import itertools as it
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable


axesdict = collections.OrderedDict([('x', 0), ('y', 1), ('z', 2)])
raxesdict = {value: key for key, value in axesdict.items()}
colormap = [0x3498db, 0xe74c3c, 0x27ae60, 0xf1c40f, 0x8e44ad, 0xecf0f1]


def array2tuple(array):
    return tuple(array.tolist())


def plane_info(*args, **kwargs):
    info = dict()
    # The plane is defined with: planeaxis and point. They are
    # extracted from *args and *kwargs.
    if args and not kwargs:
        if len(args) != 1:
            msg = 'Only one arg can be passed.'
            raise ValueError(msg)

        # Only planeaxis is provided via args and the point will be
        # defined later as a centre of the sample.
        planeaxis = args[0]
        point = None
    elif kwargs and not args:
        if len(kwargs.keys()) != 1:
            msg = 'Only one kwarg can be passed.'
            raise ValueError(msg)

        # Both planeaxis and point are provided via kwargs.
        planeaxis = list(kwargs.keys())[0]
        point = list(kwargs.values())[0]
    else:
        msg = 'Either one arg or one kwarg can be passed.'
        raise ValueError(msg)

    if planeaxis not in axesdict.keys():
        msg = f'Plane axis must be one of {axesdict.keys()}.'
        raise ValueError(msg)

    info['planeaxis'] = axesdict[planeaxis]
    info['point'] = point

    # Get indices of in-plane axes.
    axes = tuple(filter(lambda val: val != info['planeaxis'],
                        axesdict.values()))
    info['axis1'], info['axis2'] = axes

    return info


def plot_line(ax, p1, p2, *args, **kwargs):
    """Plot a line between points p1 and p2 on axis ax."""
    ax.plot(*zip(p1, p2), *args, **kwargs)


def plot_box(ax, p1, p2, *args, **kwargs):
    """Plots a cube that spans between p1 and p2 on the given axis.

    Args:
      ax (matplolib axis): matplolib axis object
      p1 (tuple, list, np.ndarray): First cube point
        p1 is of length 3 (xmin, ymin, zmax).
      p2 (tuple, list, np.ndarray): Second cube point
        p2 is of length 3 (xmin, ymin, zmax).
      color (str): matplotlib color string
      linewidth (Real): matplotlib linewidth parameter

    """
    x1, y1, z1 = p1
    x2, y2, z2 = p2

    plot_line(ax, (x1, y1, z1), (x2, y1, z1), *args, **kwargs)
    plot_line(ax, (x1, y2, z1), (x2, y2, z1), *args, **kwargs)
    plot_line(ax, (x1, y1, z2), (x2, y1, z2), *args, **kwargs)
    plot_line(ax, (x1, y2, z2), (x2, y2, z2), *args, **kwargs)

    plot_line(ax, (x1, y1, z1), (x1, y2, z1), *args, **kwargs)
    plot_line(ax, (x2, y1, z1), (x2, y2, z1), *args, **kwargs)
    plot_line(ax, (x1, y1, z2), (x1, y2, z2), *args, **kwargs)
    plot_line(ax, (x2, y1, z2), (x2, y2, z2), *args, **kwargs)

    plot_line(ax, (x1, y1, z1), (x1, y1, z2), *args, **kwargs)
    plot_line(ax, (x2, y1, z1), (x2, y1, z2), *args, **kwargs)
    plot_line(ax, (x1, y2, z1), (x1, y2, z2), *args, **kwargs)
    plot_line(ax, (x2, y2, z1), (x2, y2, z2), *args, **kwargs)


def as_array(mesh, dim, val):
    array = np.empty(mesh.n + (dim,))
    if isinstance(val, numbers.Real) and (dim == 1 or val == 0):
        # The array for a scalar field with numbers.Real value or any
        # field with zero value.
        array.fill(val)
    elif isinstance(val, (tuple, list, np.ndarray)) and len(val) == dim:
        array[..., :] = val
    elif isinstance(val, np.ndarray) and val.shape == array.shape:
        array = val
    elif callable(val):
        for index, point in zip(mesh.indices, mesh.coordinates):
            array[index] = val(point)
    else:
        msg = (f'Unsupported type(val)={type(val)} '
               'or invalid value dimensions.')
        raise ValueError(msg)
    return array


def voxels(plot_array, pmin, pmax, colormap, outlines=False,
           plot=None, **kwargs):
    plot_array = plot_array.astype(np.uint8)  # to avoid k3d warning
    if plot is None:
        plot = k3d.plot()
        plot.display()

    xmin, ymin, zmin = pmin
    xmax, ymax, zmax = pmax
    bounds = [xmin, xmax, ymin, ymax, zmin, zmax]

    plot += k3d.voxels(plot_array,
                       color_map=colormap,
                       bounds=bounds,
                       outlines=outlines,
                       **kwargs)

    return plot


def points(plot_array, point_size=0.1, color=0x99bbff, plot=None, **kwargs):
    plot_array = plot_array.astype(np.float32)  # to avoid k3d warning

    if plot is None:
        plot = k3d.plot()
        plot.display()
    plot += k3d.points(plot_array, point_size=point_size, color=color,
                       **kwargs)

    return plot


def vectors(coordinates, vectors, colors=[], plot=None, **kwargs):
    coordinates = coordinates.astype(np.float32)  # to avoid k3d warning
    vectors = vectors.astype(np.float32)  # to avoid k3d warning

    if plot is None:
        plot = k3d.plot()
        plot.display()
    plot += k3d.vectors(coordinates, vectors, colors=colors, **kwargs)

    return plot


def num2hexcolor(n, cmap):
    return int(matplotlib.colors.rgb2hex(cmap(n)[:3])[1:], 16)



import os
import types
import random
import pytest
import itertools
import numpy as np
import discretisedfield as df
import matplotlib.pyplot as plt
import discretisedfield.tests as dft


def check_field(field):
    assert isinstance(field.mesh, df.Mesh)
    assert isinstance(field.dim, int)
    assert isinstance(field.name, str)
    assert isinstance(field.array, np.ndarray)
    assert field.array.shape[-1] == field.dim
    assert isinstance(field.norm.array, np.ndarray)
    assert field.norm.array.shape[-1] == 1


class TestField:
    def setup(self):
        # Create meshes using valid arguments from TestMesh.
        tm = dft.TestMesh()
        tm.setup()
        self.meshes = []
        for p1, p2, n, cell in tm.valid_args:
            mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell)
            self.meshes.append(mesh)

        # Create lists of field values.
        self.consts = [0, -5., np.pi, 1e-15, 1.2e12, random.random()]
        self.iters = [(0, 0, 1),
                      (0, -5.1, np.pi),
                      [70, 1e15, 2*np.pi],
                      [5, random.random(), np.pi],
                      np.array([4, -1, 3.7]),
                      np.array([2.1, 0.0, -5*random.random()])]
        self.sfuncs = [lambda c: 1,
                       lambda c: -2.4,
                       lambda c: -6.4e-15,
                       lambda c: c[0] + c[1] + c[2] + 1,
                       lambda c: (c[0]-1)**2 - c[1]+7 + c[2]*0.1,
                       lambda c: np.sin(c[0]) + np.cos(c[1]) - np.sin(2*c[2])]

        self.vfuncs = [lambda c: (1, 2, 0),
                       lambda c: (-2.4, 1e-3, 9),
                       lambda c: (c[0], c[1], c[2] + 100),
                       lambda c: (c[0]+c[2]+10, c[1], c[2]+1),
                       lambda c: (c[0]-1, c[1]+70, c[2]*0.1),
                       lambda c: (np.sin(c[0]), np.cos(c[1]), -np.sin(2*c[2]))]

        # Create a field for plotting
        mesh = df.Mesh(p1=(-5, -5, -5), p2=(5, 5, 5), cell=(1, 1, 1))
        self.pf = df.Field(mesh, dim=3, value=(0, 0, 2))

        def normfun(pos):
            x, y, z = pos
            if x**2 + y**2 <= 5**2:
                return 1
            else:
                return 0

        self.pf.norm = normfun

    def test_init(self):
        p1 = (0, 0, 0)
        p2 = (5, 10, 15)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        dim = 2
        name = 'test_field'
        value = [1, 2]
        f = df.Field(mesh, dim=dim, value=value, name=name)

        check_field(f)
        assert f.name == name
        assert f.array.shape == (5, 10, 15, 2)
        assert np.all(f.array[..., 0] == value[0])
        assert np.all(f.array[..., 1] == value[1])
        assert f.norm.array.shape == (5, 10, 15, 1)

    def test_valid_init(self):
        for mesh in self.meshes:
            for value in self.consts:
                f = df.Field(mesh, dim=1, value=value)
                check_field(f)
                assert f.value == value
                assert np.all(f.array == value)
                assert f(f.mesh.random_point()) == value
                assert f(f.mesh.centre) == value
            for value in self.sfuncs:
                f = df.Field(mesh, dim=1, value=value)
                check_field(f)
            for value in self.iters:
                f = df.Field(mesh, dim=3, value=value)
                check_field(f)
                assert np.equal(f.value, value).all()
                assert np.equal(f(f.mesh.random_point()), value).all()
                assert np.equal(f(f.mesh.centre), value).all()
            for value in self.vfuncs:
                f = df.Field(mesh, dim=3, value=value)
                check_field(f)

    def test_invalid_init(self):
        with pytest.raises(TypeError):
            mesh = 'wrong_mesh_string'
            f = df.Field(mesh, dim=1)

        for mesh in self.meshes:
            with pytest.raises(TypeError):
                f = df.Field(mesh, dim='wrong_dim')

    def test_set_with_ndarray(self):
        for mesh in self.meshes:
            f = df.Field(mesh, dim=3)
            value = np.zeros(f.mesh.n + (f.dim,))
            f.value = value

            check_field(f)
            assert isinstance(f.value, np.ndarray)
            assert np.equal(f.array, 0).all()

            with pytest.raises(ValueError) as excinfo:
                f.array = (1, 2, 3)
            assert 'Unsupported' in str(excinfo.value)

    def test_set_with_function(self):
        for mesh in self.meshes:
            for func in self.sfuncs:
                f = df.Field(mesh, dim=1, value=func)
                for j in range(10):
                    c = f.mesh.random_point()
                    c = f.mesh.index2point(f.mesh.point2index(c))
                    assert f(c) == func(c)

        for mesh in self.meshes:
            for func in self.vfuncs:
                f = df.Field(mesh, dim=3, value=func)
                for j in range(10):
                    c = f.mesh.random_point()
                    c = f.mesh.index2point(f.mesh.point2index(c))
                    assert np.all(f(c) == func(c))

    def test_set_exception(self):
        for mesh in self.meshes:
            f = df.Field(mesh)
            with pytest.raises(ValueError):
                f.value = 'string'
            with pytest.raises(ValueError):
                f.value = 1+2j

    def test_repr(self):
        for mesh in self.meshes:
            f = df.Field(mesh, dim=1)
            check_field(f)
            rstr = repr(f)
            assert isinstance(rstr, str)
            assert 'mesh=' in rstr
            assert 'dim=1' in rstr
            assert 'name=' in rstr

            f = df.Field(mesh, dim=3)
            check_field(f)
            rstr = repr(f)
            assert isinstance(rstr, str)
            assert 'mesh=' in rstr
            assert 'dim=3' in rstr
            assert 'name=' in rstr

    def test_value_is_not_preserved(self):
        for mesh in self.meshes:
            f = df.Field(mesh, dim=3)
            f.value = (1, 1, 1)

            assert f.value == (1, 1, 1)

            f.array[0, 0, 0, 0] = 3
            assert isinstance(f.value, np.ndarray)

    def test_norm(self):
        mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 10), cell=(1, 1, 1))
        f = df.Field(mesh, dim=3, value=(2, 2, 2))

        assert np.all(f.norm.value == 2*np.sqrt(3))
        assert np.all(f.norm.array == 2*np.sqrt(3))
        assert np.all(f.array == 2)

        f.norm = 1
        assert np.all(f.norm.value == 1)
        assert np.all(f.norm.array == 1)
        assert np.all(f.array == 1/np.sqrt(3))

        f.array[0, 0, 0, 0] = 3
        assert isinstance(f.norm.value, np.ndarray)
        assert not np.all(f.norm.value == 1)

        for mesh in self.meshes:
            for value in self.iters + self.vfuncs:
                for norm_value in [1, 2.1, 50, 1e-3, np.pi]:
                    f = df.Field(mesh, dim=3, value=value, norm=norm_value)

                    # Compute norm.
                    norm = f.array[..., 0]**2
                    norm += f.array[..., 1]**2
                    norm += f.array[..., 2]**2
                    norm = np.sqrt(norm)

                    assert norm.shape == f.mesh.n
                    assert f.norm.array.shape == f.mesh.n + (1,)
                    assert np.all(abs(norm - norm_value) < 1e-12)

    def test_norm_is_not_preserved(self):
        for mesh in self.meshes:
            f = df.Field(mesh, dim=3)
            f.value = (0, 3, 0)
            f.norm = 1
            assert np.all(f.norm.array == 1)

            f.value = (0, 2, 0)
            assert np.all(f.norm.value != 1)
            assert np.all(f.norm.array == 2)

    def test_norm_scalar_field_exception(self):
        for mesh in self.meshes:
            for value in self.consts + self.sfuncs:
                for norm in [1, 2.1, 50, 1e-3, np.pi]:
                    f = df.Field(mesh, dim=1, value=value)
                    with pytest.raises(ValueError):
                        f.norm = norm

    def test_norm_zero_field_exception(self):
        for mesh in self.meshes:
            f = df.Field(mesh, dim=3, value=(0, 0, 0))
            with pytest.raises(ValueError):
                f.norm = 1

    def test_average(self):
        value = -1e-3 + np.pi
        tol = 1e-12
        for mesh in self.meshes:
            f = df.Field(mesh, dim=1, value=value)

            assert isinstance(f.average, tuple)
            assert len(f.average) == 1
            assert abs(f.average[0] - value) < tol

        value = np.array([1.1, -2e-3, np.pi])
        tol = 1e-12
        for mesh in self.meshes:
            f = df.Field(mesh, dim=3, value=value)

            assert isinstance(f.average, tuple)
            assert len(f.average) == 3
            assert np.allclose(f.average, value)

    def test_field_component(self):
        for mesh in self.meshes:
            f = df.Field(mesh, dim=3, value=(1, 2, 3))
            assert isinstance(f.x, df.Field)
            assert f.x.dim == 1
            assert np.all(f.x.array == 1)
            assert isinstance(f.y, df.Field)
            assert f.y.dim == 1
            assert np.all(f.y.array == 2)
            assert isinstance(f.z, df.Field)
            assert f.z.dim == 1
            assert np.all(f.z.array == 3)

            f = df.Field(mesh, dim=2, value=(1, 2))
            assert isinstance(f.x, df.Field)
            assert f.x.dim == 1
            assert np.all(f.x.array == 1)
            assert isinstance(f.y, df.Field)
            assert f.y.dim == 1
            assert np.all(f.y.array == 2)

            f = df.Field(mesh, dim=1, value=1)
            with pytest.raises(AttributeError):
                assert f.x.dim == 1

    def test_get_attribute_exception(self):
        for mesh in self.meshes:
            f = df.Field(mesh)
            with pytest.raises(AttributeError) as excinfo:
                f.__getattr__('nonexisting_attribute')
            assert 'has no attribute' in str(excinfo.value)

    def test_dir(self):
        for mesh in self.meshes:
            f = df.Field(mesh, value=(5, 6, -9))
            assert 'x' in f.__dir__()
            assert 'y' in f.__dir__()
            assert 'z' in f.__dir__()

            f = df.Field(mesh, dim=1, value=1)
            assert 'x' not in f.__dir__()
            assert 'y' not in f.__dir__()
            assert 'z' not in f.__dir__()

    def test_line(self):
        mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 10), n=(10, 10, 10))
        f = df.Field(mesh, value=(1, 2, 3))

        line = f.line(p1=(0, 0, 0), p2=(5, 5, 5), n=20)
        assert isinstance(line, types.GeneratorType)
        p, v = zip(*list(line))

        assert len(p) == 20
        assert len(v) == 20
        assert p[0] == (0, 0, 0)
        assert p[-1] == (5, 5, 5)
        assert v[0] == (1, 2, 3)
        assert v[-1] == (1, 2, 3)

    def test_plane(self):
        for mesh in self.meshes:
            f = df.Field(mesh, dim=1, value=3)
            plane = f.plane('x', n=(3, 3))
            assert isinstance(plane, df.Field)

            p, v = zip(*list(plane))
            assert len(p) == 9
            assert len(v) == 9

    def test_writevtk(self):
        vtkfilename = 'test_file.ovf'
        mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 12, 13), cell=(1, 1, 1))
        f = df.Field(mesh, dim=3, value=(1, 2, -5))
        f.write(vtkfilename)
        os.remove(vtkfilename)

    def test_write_wrong_file(self):
        wrongfilename = 'test_file.jpg'
        mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 12, 13), cell=(1, 1, 1))
        f = df.Field(mesh, dim=3, value=(1, 2, -5))
        with pytest.raises(ValueError) as excinfo:
            f.write(wrongfilename)
        assert 'Allowed extensions' in str(excinfo.value)

    def test_wrong_dim_field(self):
        ovffilename = 'test_file.ovf'
        mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 12, 13), cell=(1, 1, 1))
        f = df.Field(mesh, dim=2, value=(1, 2))
        with pytest.raises(TypeError) as excinfo:
            f.write(ovffilename)
        assert 'Cannot write' in str(excinfo.value)

    def test_writeovf(self):
        representations = ['txt', 'bin4', 'bin8']
        tolerance = {'txt': 0, 'bin4': 1e-6, 'bin8': 1e-12}
        filename = 'test_file.ovf'
        mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 12, 13), cell=(1, 1, 1))

        for dim, value in [(1, -1.23), (3, (1e-3 + np.pi, -5, 6))]:
            f_out = df.Field(mesh, dim=dim, value=value)
            for rep in representations:
                f_out.write(filename, representation=rep)
                f_in = df.Field.fromfile(filename)
                assert mesh.p1 == f_in.mesh.p1
                assert mesh.p2 == f_in.mesh.p2
                assert mesh.cell == f_in.mesh.cell
                np.testing.assert_allclose(f_out.array, f_in.array,
                                           rtol=tolerance[rep])

        os.remove(filename)

    def test_read_mumax3_ovffile(self):
        """Test reading a file created by mumax, with 4-byte binary data."""
        # Output file has been produced with Mumax ~3.1.0
        # (fd3a50233f0a6625086d390) using this script:
        #
        # SetGridsize(128, 32, 1)
        # SetCellsize(500e-9/128, 125e-9/32, 3e-9)
        #
        # Msat  = 800e3
        # Aex   = 13e-12
        # alpha = 0.02
        #
        # m = uniform(1, .1, 0)
        # save(m)

        # debug: which folder are we in? (use "py.test -l" to display
        # local variables on fail)
        cwd = os.getcwd()
        # -> tests are run in module base directory, when tests are
        #    run from discretisedfield.test()

        # here is our test-data from mumax3:
        filenames = ["mumax-output-linux.ovf", "mumax-output-win.ovf"]
        test_sample_dirname = os.path.join(os.path.dirname(__file__),
                                           'test_sample/')
        for f in filenames:
            path = os.path.join(test_sample_dirname, f)

            f = df.Field.fromfile(path)

            # comparison with human readable part of file
            assert f.dim == 3
            assert f.mesh.ntotal == 4096
            assert f.mesh.pmin == (0., 0., 0.)
            assert f.mesh.pmax == (5e-07, 1.25e-07, 3e-09)
            assert f.array.shape == (128, 32, 1, 3)

            # comparison with vector field (we know from the script
            # shown above)
            #
            # m vector in mumax (uses 4 bytes)
            m = np.array([1, 0.1, 0], dtype=np.float32)

            # needs to be normalised
            norm = sum(m**2)**0.5
            v = m / norm

            # expect the x-component to be v[0]
            np.alltrue(f.array[:, :, :, 0] == v[0])

            # same for y and z
            np.alltrue(f.array[:, :, :, 1] == v[1])
            np.alltrue(f.array[:, :, :, 2] == v[2])

            # expect each vector to be v
            assert np.max(np.abs(f.array[:, :, :] - m / norm)) == 0.

    def test_mpl(self):
        with pytest.raises(ValueError) as excinfo:
            self.pf.mpl()
        assert 'Only sliced field' in str(excinfo.value)

        self.pf.plane('x', n=(3, 4)).mpl()
        self.pf.z.plane('x', n=(3, 4)).mpl()

    def test_imshow(self):
        fig = plt.figure()
        ax = fig.add_subplot(111)

        with pytest.raises(ValueError) as excinfo:
            self.pf.imshow(ax=ax)
        assert 'Only sliced field' in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            self.pf.plane('z').imshow(ax=ax)
        assert 'Only scalar' in str(excinfo.value)

        self.pf.x.plane('z', n=(3, 4)).imshow(ax=ax)
        self.pf.x.plane('x', n=(3, 4)).imshow(ax=ax, norm_field=self.pf.norm)

    def test_quiver(self):
        fig = plt.figure()
        ax = fig.add_subplot(111)

        with pytest.raises(ValueError) as excinfo:
            self.pf.quiver(ax=ax)
        assert 'Only sliced field' in str(excinfo.value)

        with pytest.raises(ValueError) as excinfo:
            self.pf.x.plane('y').quiver(ax=ax)
        assert 'Only three-dimensional' in str(excinfo.value)

        self.pf.plane('x', n=(3, 4)).quiver(ax=ax)
        self.pf.plane('x', n=(3, 4)).quiver(ax=ax, color_field=self.pf.y)

    def test_colorbar(self):
        fig = plt.figure()
        ax = fig.add_subplot(111)

        coloredplot = self.pf.x.plane('x', n=(3, 4)).imshow(ax=ax)
        self.pf.colorbar(ax=ax, coloredplot=coloredplot)

    def test_k3d_nonzero(self):
        with pytest.raises(ValueError) as excinfo:
            self.pf.k3d_nonzero()
        assert 'Only scalar' in str(excinfo.value)

        self.pf.norm.k3d_nonzero()

    def test_k3d_voxels(self):
        with pytest.raises(ValueError) as excinfo:
            self.pf.k3d_voxels()
        assert 'Only scalar' in str(excinfo.value)

        self.pf.x.k3d_voxels()
        self.pf.x.k3d_voxels(norm_field=self.pf.norm)

    def test_k3d_vectors(self):
        with pytest.raises(ValueError) as excinfo:
            self.pf.x.k3d_vectors()
        assert 'Only three-dimensional' in str(excinfo.value)

        self.pf.k3d_vectors()
        self.pf.k3d_vectors(color_field=self.pf.z)
        self.pf.k3d_vectors(points=False)



import pyvtk
import struct
import matplotlib
import numpy as np
import mpl_toolkits.axes_grid1
import discretisedfield as df
import ubermagutil.typesystem as ts
import discretisedfield.util as dfu
import matplotlib.pyplot as plt


@ts.typesystem(mesh=ts.Typed(expected_type=df.Mesh),
               dim=ts.Scalar(expected_type=int, unsigned=True, const=True),
               name=ts.Name(const=True))
class Field:
    """Finite difference field.

    This class defines a finite difference field and enables certain
    operations for its analysis and visualisation. The field is
    defined on a finite difference mesh (`discretisedfield.Mesh`).

    Parameters
    ----------
    mesh : discretisedfield.Mesh
        Finite difference rectangular mesh.
    dim : int, optional
        Dimension of the field value. For instance, if `dim=3` the
        field is a three-dimensional vector field and for `dim=1`
        the field is a scalar field. Defaults to `dim=3`.
    value : array_like, callable, optional
        Please refer to the `value` property:
        :py:func:`~discretisedfield.Field.value`. Defaults to 0,
        meaning that if the value is not provided in the
        initialisation process, "zero-field" will be defined.
    norm : numbers.Real, callable, optional
        Please refer to the `norm` property:
        :py:func:`~discretisedfield.Field.norm`. Defaults to `None`
        (`norm=None` defines no norm).
    name : str, optional
        Field name (defaults to `'field'`). The field name must be a
        valid Python variable name string. More specifically, it must
        not contain spaces, or start with underscore or numeric
        character.

    Examples
    --------
    1. Creating a uniform three-dimensional vector field on a
    nano-sized thin film.

    >>> import discretisedfield as df
    ...
    >>> p1 = (-50e-9, -25e-9, 0)
    >>> p2 = (50e-9, 25e-9, 5e-9)
    >>> cell = (1e-9, 1e-9, 0.1e-9)
    >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
    ...
    >>> dim = 3
    >>> value = (0, 0, 1)
    >>> field = df.Field(mesh=mesh, dim=dim, value=value)
    >>> field
    Field(mesh=...)

    2. Creating a scalar field.

    >>> import discretisedfield as df
    ...
    >>> p1 = (-10, -10, -10)
    >>> p2 = (10, 10, 10)
    >>> n = (1, 1, 1)
    >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
    >>> dim = 1
    >>> value = 3.14
    >>> field = df.Field(mesh=mesh, dim=dim, value=value)
    >>> field
    Field(mesh=...)

    .. seealso:: :py:func:`~discretisedfield.Mesh`

    """
    def __init__(self, mesh, dim=3, value=0, norm=None, name='field'):
        self.mesh = mesh
        self.dim = dim
        self.value = value
        self.norm = norm
        self.name = name

    @property
    def value(self):
        """Field value representation.

        This propertry returns a representation of the field value if
        it exists. Otherwise, the `numpy.ndarray` containing all
        values from the field is returned.

        Parameters
        ----------
        value : 0, array_like, callable
            For scalar fields (`dim=1`) `numbers.Real` values are
            allowed. In the case of vector fields, "array_like" (list,
            tuple, numpy.ndarray) value with length equal to `dim`
            should be used. Finally, the value can also be a callable
            (e.g. Python function or another field), which for every
            coordinate in the mesh returns a valid value. If
            `value=0`, all values in the field will be set to zero
            independent of the field dimension.

        Returns
        -------
        array_like, callable, numbers.Real
            The value used (representation) for setting the field is
            returned. However, if the actual value of the field does
            not correspond to the initially used value anymore, a
            `numpy.ndarray` is returned containing all field values.

        Raises
        ------
        ValueError
            If unsupported type is passed

        Examples
        --------
        1. Different ways of setting and getting the field value.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> value = (0, 0, 1)
        >>> # if value is not specified, zero-field is defined
        >>> field = df.Field(mesh=mesh, dim=3)
        >>> field.value
        0
        >>> field.value = (0, 0, 1)
        >>> field.value
        (0, 0, 1)
        >>> # Setting the field value using a Python function (callable).
        >>> def value_function(pos):
        ...     x, y, z = pos
        ...     if x <= 1:
        ...         return (0, 0, 1)
        ...     else:
        ...         return (0, 0, -1)
        >>> field.value = value_function
        >>> field.value
        <function value_function at ...>
        >>> # We now change the value of a single cell so that the
        >>> # representation used for initialising field is not valid
        >>> # anymore.
        >>> field.array[0, 0, 0, :] = (0, 0, 0)
        >>> field.value
        array(...)

        .. seealso:: :py:func:`~discretisedfield.Field.array`

        """
        value_array = dfu.as_array(self.mesh, self.dim, self._value)
        if np.array_equal(self.array, value_array):
            return self._value
        else:
            return self.array

    @value.setter
    def value(self, val):
        self._value = val
        self.array = dfu.as_array(self.mesh, self.dim, val)

    @property
    def array(self):
        """Numpy array of a field value.

        `array` has shape of `(self.mesh.n[0], self.mesh.n[1],
        self.mesh.n[2], dim)`.

        Parameters
        ----------
        array : numpy.ndarray
            Numpy array with dimensions `(self.mesh.n[0],
            self.mesh.n[1], self.mesh.n[2], dim)`

        Returns
        -------
        numpy.ndarray
            Field values array.

        Raises
        ------
        ValueError
            If setting the array with wrong type, shape, or value.

        Examples
        --------
        1. Accessing and setting the field array.

        >>> import discretisedfield as df
        >>> import numpy as np
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (1, 1, 1)
        >>> cell = (0.5, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> value = (0, 0, 1)
        >>> field = df.Field(mesh=mesh, dim=3, value=value)
        >>> field.array
        array(...)
        >>> field.array.shape
        (2, 1, 1, 3)
        >>> field.array = np.ones(field.array.shape)
        >>> field.array
        array(...)

        .. seealso:: :py:func:`~discretisedfield.Field.value`

        """
        return self._array

    @array.setter
    def array(self, val):
        if isinstance(val, np.ndarray) and \
           val.shape == self.mesh.n + (self.dim,):
            self._array = val
        else:
            msg = (f'Unsupported type(val)={type(val)} '
                   'or invalid value dimensions.')
            raise ValueError(msg)

    @property
    def norm(self):
        """Norm of a field.

        This property computes the norm of the field and returns it as
        a `discretisedfield.Field` object with `dim=1`. Norm of a
        scalar field cannot be set and `ValueError` is raised.

        Parameters
        ----------
        numbers.Real, numpy.ndarray
            Norm value

        Returns
        -------
        discretisedfield.Field
            Scalar field with norm values.

        Raises
        ------
        ValueError
            If setting the norm with wrong type, shape, or value. In
            addition, if the field is scalar (dim=1) or it contains
            zero vector values.

        Examples
        --------
        1. Manipulating the field norm

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (1, 1, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> field = df.Field(mesh=mesh, dim=3, value=(0, 0, 1))
        >>> field.norm
        Field(...)
        >>> field.norm = 2
        >>> field.norm
        Field(...)
        >>> field.value = (1, 0, 0)
        >>> field.norm.array
        array([[[[1.]]]])

        """
        current_norm = np.linalg.norm(self.array, axis=-1)[..., None]
        return Field(self.mesh, dim=1, value=current_norm, name='norm')

    @norm.setter
    def norm(self, val):
        if val is not None:
            if self.dim == 1:
                msg = f'Cannot set norm for field with dim={self.dim}.'
                raise ValueError(msg)

            if not np.all(self.norm.array):
                msg = 'Cannot normalise field with zero values.'
                raise ValueError(msg)

            self.array /= self.norm.array  # normalise to 1
            self.array *= dfu.as_array(self.mesh, dim=1, val=val)

    @property
    def average(self):
        """Field average.

        It computes the average of the field over the entire volume of
        the mesh.

        Returns
        -------
        tuple
            Field average tuple whose length equals to the field's
            dimension.

        Examples
        --------
        1. Computing the vector field average.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (5, 5, 5)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> field1 = df.Field(mesh=mesh, dim=3, value=(0, 0, 1))
        >>> field1.average
        (0.0, 0.0, 1.0)
        >>> field2 = df.Field(mesh=mesh, dim=1, value=55)
        >>> field2.average
        (55.0,)

        """
        return tuple(self.array.mean(axis=(0, 1, 2)))

    def __repr__(self):
        """Field representation string.

        This method returns the string that can ideally be copied in
        another Python script so that exactly the same field object
        could be defined. However, this is usually not the case due to
        complex values used.

        Returns
        -------
        str
            Field representation string.

        Example
        -------
        1. Getting field representation string.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> field = df.Field(mesh, dim=1, value=1)
        >>> repr(field)
        "Field(mesh=...)"

        """
        return (f'Field(mesh={repr(self.mesh)}, '
                f'dim={self.dim}, name=\'{self.name}\')')

    def __call__(self, point):
        """Sample the field at `point`.

        It returns the value of the discreatisation cell `point`
        belongs to. It always returns a tuple, whose length is the
        same as the dimension of the field.

        Parameters
        ----------
        point : (3,) array_like
            The mesh point coordinate :math:`(p_{x}, p_{y}, p_{z})`.

        Returns
        -------
        tuple
            A tuple, whose length is the same as the dimension of the
            field.

        Example
        -------
        1. Sampling the field value

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (20, 20, 20)
        >>> n = (20, 20, 20)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, dim=3, value=(1, 3, 4))
        >>> point = (10, 2, 3)
        >>> field(point)
        (1.0, 3.0, 4.0)

        """
        value = self.array[self.mesh.point2index(point)]
        if self.dim > 1:
            value = tuple(value)
        return value

    def __getattr__(self, name):
        """Extracting the component of the vector field.

        If `'x'`, `'y'`, or `'z'` is accessed, a new scalar field of
        that component will be returned. This method is effective for
        vector fields with dimension 2 or 3.

        Returns
        -------
        discretisedfield.Field
            Scalar field with vector field component values.

        Examples
        --------
        1. Accessing the vector field components.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 2)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> field = df.Field(mesh=mesh, dim=3, value=(0, 0, 1))
        >>> field.x
        Field(...)
        >>> field.y
        Field(...)
        >>> field.z
        Field(...)
        >>> field.z.dim
        1

        """
        if name in list(dfu.axesdict.keys())[:self.dim] and 1 < self.dim <= 3:
            # Components x, y, and z make sense only for vector fields
            # with typical dimensions 2 and 3.
            component_array = self.array[..., dfu.axesdict[name]][..., None]
            fieldname = f'{self.name}-{name}'.format(self.name, name)
            return Field(mesh=self.mesh, dim=1,
                         value=component_array, name=fieldname)
        else:
            msg = f'{type(self).__name__} object has no attribute {name}.'
            raise AttributeError(msg.format(type(self).__name__, name))

    def __dir__(self):
        """Extension of the tab-completion list.

        Adds `'x'`, `'y'`, and `'z'`, depending on the dimension of
        the field, to the tab-completion list. This is effective in
        IPython or Jupyter notebook environment.

        """
        if 1 < self.dim <= 3:
            extension = list(dfu.axesdict.keys())[:self.dim]
        else:
            extension = []
        return list(self.__dict__.keys()) + extension

    def __iter__(self):
        """Generator yielding coordinates and values of all field cells.

        The discretisation cell coordinate corresponds to the cell
        centre point.

        Yields
        ------
        tuple (2,)
            The first value is the mesh cell coordinates (`px`, `py`,
            `pz`), whereas the second one is the field value.

        Examples
        --------
        1. Iterating through the field coordinates and values

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> field = df.Field(mesh, dim=3, value=(0, 0, 1))
        >>> for coord, value in field:
        ...     print (coord, value)
        (0.5, 0.5, 0.5) (0.0, 0.0, 1.0)
        (1.5, 0.5, 0.5) (0.0, 0.0, 1.0)
        (0.5, 1.5, 0.5) (0.0, 0.0, 1.0)
        (1.5, 1.5, 0.5) (0.0, 0.0, 1.0)

        .. seealso:: :py:func:`~discretisedfield.Mesh.indices`

        """
        for point in self.mesh.coordinates:
            yield point, self.__call__(point)

    def line(self, p1, p2, n=100):
        """Sampling the field along the line.

        Given two points :math:`p_{1}` and :math:`p_{2}`, :math:`n`
        position coordinates are generated and the corresponding field
        values.

        .. math::

           \\mathbf{r}_{i} = i\\frac{\\mathbf{p}_{2} -
           \\mathbf{p}_{1}}{n-1}

        Parameters
        ----------
        p1, p2 : (3,) array_like
            Two points between which the line is generated.
        n : int
            Number of points on the line.

        Yields
        ------
        tuple
            The first element is the coordinate of the point on the
            line, whereas the second one is the value of the field.

        Raises
        ------
        ValueError
            If `p1` or `p2` is outside the mesh domain.

        Examples
        --------
        1. Sampling the field along the line.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 2)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> field = df.Field(mesh, dim=2, value=(0, 3))
        >>> for coord, value in field.line(p1=(0, 0, 0), p2=(2, 0, 0), n=3):
        ...     print(coord, value)
        (0.0, 0.0, 0.0) (0.0, 3.0)
        (1.0, 0.0, 0.0) (0.0, 3.0)
        (2.0, 0.0, 0.0) (0.0, 3.0)

        """
        for point in self.mesh.line(p1=p1, p2=p2, n=n):
            yield point, self.__call__(point)

    def plane(self, *args, n=None, **kwargs):
        """Slices the field with a plane.

        If one of the axes (`'x'`, `'y'`, or `'z'`) is passed as a
        string, a plane perpendicular to that axis is generated which
        intersects the field at its centre. Alternatively, if a keyword
        argument is passed (e.g. `x=1`), a plane perpendicular to the
        x-axis and intersecting it at x=1 is generated. The number of
        points in two dimensions on the plane can be defined using `n`
        (e.g. `n=(10, 15)`). Using the generated plane, a new
        "two-dimensional" field is created and returned.

        Parameters
        ----------
        n : tuple of length 2
            The number of points on the plane in two dimensions

        Returns
        ------
        discretisedfield.Field
            A field obtained as an intersection of mesh and the plane.

        Example
        -------
        1. Intersecting the field with a plane.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 2)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> field = df.Field(mesh, dim=3)
        >>> field.plane(y=1)
        Field(mesh=...)

        """
        plane_mesh = self.mesh.plane(*args, n=n, **kwargs)
        return self.__class__(plane_mesh, dim=self.dim, value=self)

    def write(self, filename, representation='txt'):
        """Write the field in .ovf, .omf, .ohf, or vtk format.

        If the extension of the `filename` is `.vtk`, a VTK file is
        written
        (:py:func:`~discretisedfield.Field._writevtk`). Otherwise, for
        `.ovf`, `.omf`, or `.ohf` extensions, an OOMMF file is written
        (:py:func:`~discretisedfield.Field._writeovf`). The
        representation (`bin4`, 'bin8', or 'txt') is passed using
        `representation` argument.

        Parameters
        ----------
        filename : str
            Name of the file written. It depends on its extension the
            format it is going to be written as.
        representation : str
            In the case of OOMMF files (`.ovf`, `.omf`, or `.ohf`),
            representation can be specified (`bin4`, `bin8`, or
            `txt`). Defaults to 'txt'.

        Example
        -------
        1. Write an .omf file and delete it from the disk

        >>> import os
        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, -5)
        >>> p2 = (5, 15, 15)
        >>> n = (5, 15, 20)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, value=(5, 6, 7))
        >>> filename = 'mytestfile.omf'
        >>> field.write(filename)  # write the file
        >>> os.remove(filename)  # delete the file

        .. seealso:: :py:func:`~discretisedfield.Field.fromfile`

        """
        if any([filename.endswith(ext) for ext in ['.omf', '.ovf', '.ohf']]):
            self._writeovf(filename, representation=representation)
        elif filename.endswith('.vtk'):
            self._writevtk(filename)
        else:
            msg = ('Allowed extensions for writing the field are '
                   '.omf, .ovf, .ohf, and .vtk.')
            raise ValueError(msg)

    def _writeovf(self, filename, representation='txt'):
        """Write the field in .ovf, .omf, or .ohf format.

        The extension of the `filename` should be `.ovf`, `.omf`, or
        `.ohf`. The representation (`bin4`, 'bin8', or 'txt') is
        passed using `representation` argument.

        Parameters
        ----------
        filename : str
            Name of the file written.
        representation : str
            Representation of the file (`bin4`, `bin8`, or
            `txt`). Defaults to 'txt'.

        Example
        -------
        1. Write an .omf file and delete it from the disk

        >>> import os
        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, -5)
        >>> p2 = (5, 15, 15)
        >>> n = (5, 15, 20)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, value=(5, 6, 7))
        >>> filename = 'mytestfile.omf'
        >>> field._writeovf(filename)  # write the file
        >>> os.remove(filename)  # delete the file

        """
        header = ['OOMMF OVF 2.0',
                  '',
                  'Segment count: 1',
                  '',
                  'Begin: Segment',
                  'Begin: Header',
                  '',
                  'Title: Field generated omf file',
                  'Desc: File generated by Field class',
                  'meshunit: m',
                  'meshtype: rectangular',
                  f'xbase: {self.mesh.pmin[0] + self.mesh.cell[0]/2}',
                  f'ybase: {self.mesh.pmin[1] + self.mesh.cell[1]/2}',
                  f'zbase: {self.mesh.pmin[2] + self.mesh.cell[2]/2}',
                  f'xnodes: {self.mesh.n[0]}',
                  f'ynodes: {self.mesh.n[1]}',
                  f'znodes: {self.mesh.n[2]}',
                  f'xstepsize: {self.mesh.cell[0]}',
                  f'ystepsize: {self.mesh.cell[1]}',
                  f'zstepsize: {self.mesh.cell[2]}',
                  f'xmin: {self.mesh.pmin[0]}',
                  f'ymin: {self.mesh.pmin[1]}',
                  f'zmin: {self.mesh.pmin[2]}',
                  f'xmax: {self.mesh.pmax[0]}',
                  f'ymax: {self.mesh.pmax[1]}',
                  f'zmax: {self.mesh.pmax[2]}',
                  f'valuedim: {self.dim}',
                  f'valuelabels: {self.name}_x {self.name}_y {self.name}_z',
                  'valueunits: A/m A/m A/m',
                  '',
                  'End: Header',
                  '']

        if representation == 'bin4':
            header.append('Begin: Data Binary 4')
            footer = ['End: Data Binary 4',
                      'End: Segment']
        elif representation == 'bin8':
            header.append('Begin: Data Binary 8')
            footer = ['End: Data Binary 8',
                      'End: Segment']
        elif representation == 'txt':
            header.append('Begin: Data Text')
            footer = ['End: Data Text',
                      'End: Segment']

        # Write header lines to the ovf file.
        f = open(filename, 'w')
        f.write(''.join(map(lambda line: f'# {line}\n', header)))
        f.close()

        binary_reps = {'bin4': (1234567.0, 'f'),
                       'bin8': (123456789012345.0, 'd')}

        if representation in binary_reps:
            # Reopen the file with binary write, appending to the end
            # of the file.
            f = open(filename, 'ab')

            # Add the 8 bit binary check value that OOMMF uses.
            packarray = [binary_reps[representation][0]]

            # Write data to the ovf file.
            for i in self.mesh.indices:
                for vi in self.array[i]:
                    packarray.append(vi)

            v_bin = struct.pack(binary_reps[representation][1]*len(packarray),
                                *packarray)
            f.write(v_bin)
            f.close()

        else:
            # Reopen the file for txt representation, appending to the
            # file.
            f = open(filename, 'a')
            for i in self.mesh.indices:
                if self.dim == 3:
                    v = [vi for vi in self.array[i]]
                elif self.dim == 1:
                    v = [self.array[i][0]]
                else:
                    msg = (f'Cannot write dim={self.dim} field.')
                    raise TypeError(msg)
                for vi in v:
                    f.write(' ' + str(vi))
                f.write('\n')
            f.close()

        # Write footer lines to OOMMF file.
        f = open(filename, 'a')
        f.write(''.join(map(lambda line: f'# {line}\n', footer)))
        f.close()

    def _writevtk(self, filename):
        """Write the field in the VTK format.

        The extension of the `filename` should be `.vtk`.

        Parameters
        ----------
        filename : str
            Name of the file written.

        Example
        -------
        1. Write a .vtk file and delete it from the disk

        >>> import os
        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, -5)
        >>> p2 = (5, 15, 15)
        >>> n = (5, 15, 20)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, value=(5, 6, 7))
        >>> filename = 'mytestfile.vtk'
        >>> field._writevtk(filename)  # write the file
        >>> os.remove(filename)  # delete the file

        """
        grid = [pmini + np.linspace(0, li, ni+1) for pmini, li, ni in
                zip(self.mesh.pmin, self.mesh.l, self.mesh.n)]

        structure = pyvtk.RectilinearGrid(*grid)
        vtkdata = pyvtk.VtkData(structure)

        vectors = [self.__call__(coord) for coord in self.mesh.coordinates]
        vtkdata.cell_data.append(pyvtk.Vectors(vectors, self.name))
        for i, component in enumerate(dfu.axesdict.keys()):
            name = f'{self.name}_{component}'
            vtkdata.cell_data.append(pyvtk.Scalars(list(zip(*vectors))[i],
                                                   name))

        vtkdata.tofile(filename)

    @classmethod
    def fromfile(cls, filename, norm=None, name='field'):
        """Read the field from .ovf, .omf, or .ohf file.

        The extension of the `filename` should be `.ovf`, `.omf`, or
        `.ohf`. If the field should be normalised, `norm` argument can
        be passed. The `name` of the field defaults to `'field'`. This
        is a `classmethod` and should be called as
        `discretisedfield.Field.fromfile('myfile.omf')`.

        Parameters
        ----------
        filename : str
            Name of the file to be read.
        norm : numbers.Real, numpy.ndarray, callable
            For details, refer to :py:func:`~discretisedfield.Field.value`.
        name : str
            Name of the field read.

        Returns
        -------
        discretisedfield.Field

        Example
        -------
        1. Read a field from the .ovf file

        >>> import os
        >>> import discretisedfield as df
        ...
        >>> ovffile = os.path.join(os.path.dirname(__file__),
        ...                        'tests', 'test_sample',
        ...                        'mumax-output-linux.ovf')
        >>> field = df.Field.fromfile(ovffile)
        >>> field
        Field(mesh=...)

        .. seealso:: :py:func:`~discretisedfield.Field.write`

        """
        mdatalist = ['xmin', 'ymin', 'zmin', 'xmax', 'ymax', 'zmax',
                     'xstepsize', 'ystepsize', 'zstepsize', 'valuedim']
        mdatadict = dict()

        try:
            with open(filename, 'r', encoding='utf-8') as ovffile:
                f = ovffile.read()
                lines = f.split('\n')

            mdatalines = filter(lambda s: s.startswith('#'), lines)
            datalines = np.loadtxt(filter(lambda s: not s.startswith('#'),
                                          lines))

            for line in mdatalines:
                for mdatum in mdatalist:
                    if mdatum in line:
                        mdatadict[mdatum] = float(line.split()[-1])
                        break

        except UnicodeDecodeError:
            with open(filename, 'rb') as ovffile:
                f = ovffile.read()
                lines = f.split(b'\n')

            mdatalines = filter(lambda s: s.startswith(bytes('#', 'utf-8')),
                                lines)

            for line in mdatalines:
                for mdatum in mdatalist:
                    if bytes(mdatum, 'utf-8') in line:
                        mdatadict[mdatum] = float(line.split()[-1])
                        break

            header = b'# Begin: Data Binary '
            data_start = f.find(header)
            header = f[data_start:data_start + len(header) + 1]

            data_start += len(b'# Begin: Data Binary 8')
            data_end = f.find(b'# End: Data Binary ')

            # ordered by length
            newlines = [b'\n\r', b'\r\n', b'\n']
            for nl in newlines:
                if f.startswith(nl, data_start):
                    data_start += len(nl)
                    break

            if b'4' in header:
                formatstr = '@f'
                checkvalue = 1234567.0
            elif b'8' in header:
                formatstr = '@d'
                checkvalue = 123456789012345.0

            listdata = list(struct.iter_unpack(formatstr,
                                               f[data_start:data_end]))
            datalines = np.array(listdata)

            if datalines[0] != checkvalue:
                # These two lines cannot be accessed via
                # tests. Therefore, they are excluded from coverage.
                msg = 'Binary Data cannot be read.'  # pragma: no cover
                raise AssertionError(msg)  # pragma: no cover

            datalines = datalines[1:]  # check value removal

        p1 = (mdatadict[key] for key in ['xmin', 'ymin', 'zmin'])
        p2 = (mdatadict[key] for key in ['xmax', 'ymax', 'zmax'])
        cell = (mdatadict[key] for key in ['xstepsize', 'ystepsize',
                                           'zstepsize'])
        dim = int(mdatadict['valuedim'])

        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

        field = df.Field(mesh, dim=dim, name=name)

        r_tuple = tuple(reversed(field.mesh.n)) + (int(mdatadict['valuedim']),)
        t_tuple = tuple(reversed(range(3))) + (3,)
        field.array = datalines.reshape(r_tuple).transpose(t_tuple)
        field.norm = norm  # Normalise if norm is passed

        return field

    def mpl(self, figsize=None):
        """Plots a field plane using matplotlib.

        Before the field can be plotted, it must be sliced with a
        plane (e.g. `field.plane(`z`)`). Otherwise, ValueError is
        raised. For vector fields, this method plots both `quiver`
        (vector) and `imshow` (scalar) plots. The `imshow` plot
        represents the value of the out-of-plane vector component and
        the `quiver` plot is not coloured. On the other hand, only
        `imshow` is plotted for scalar fields. Where the norm of the
        field is zero, no vectors are shown and those `imshow` pixels
        are not coloured. In order to use this function inside Jupyter
        notebook `%matplotlib inline` must be activated after
        `discretisedfield` is imported.

        Parameters
        ----------
        figsize : tuple, optional
            Length-2 tuple passed to the `matplotlib.figure` function.

        Raises
        ------
        ValueError
            If the field has not been sliced with a plane.

        Example
        -------
        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (100, 100, 100)
        >>> n = (10, 10, 10)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, dim=3, value=(1, 2, 0))
        >>> field.plane(z=50, n=(5, 5)).mpl()

        .. seealso:: :py:func:`~discretisedfield.Field.k3d_vectors`

        """
        if not hasattr(self.mesh, 'info'):
            msg = ('Only sliced field can be plotted using mpl. '
                   'For instance, field.plane(\'x\').mpl().')
            raise ValueError(msg)

        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)

        planeaxis = dfu.raxesdict[self.mesh.info['planeaxis']]

        if self.dim > 1:
            # Vector field has both quiver and imshow plots.
            self.quiver(ax=ax, headwidth=5)
            scfield = getattr(self, planeaxis)
            coloredplot = scfield.imshow(ax=ax, norm_field=self.norm)
        else:
            # Scalar field has only imshow.
            scfield = self
            coloredplot = scfield.imshow(ax=ax, norm_field=None)

        # Add colorbar to imshow plot.
        cbar = self.colorbar(ax, coloredplot)

        # Add labels.
        ax.set_xlabel(dfu.raxesdict[self.mesh.info['axis1']])
        ax.set_ylabel(dfu.raxesdict[self.mesh.info['axis2']])
        if self.dim > 1:
            cbar.ax.set_ylabel(planeaxis + ' component')

    def imshow(self, ax, norm_field=None, **kwargs):
        """Plots a scalar field plane using `matplotlib.pyplot.imshow`.

        Before the field can be plotted, it must be sliced with a
        plane (e.g. `field.plane(`y`)`) and field must be of dimension
        1 (scalar field). Otherwise, ValueError is raised. `imshow`
        adds the plot to `matplotlib.axes.Axes` passed via `ax`
        argument. If the scalar field plotted is extracted from a
        vector field, which has coordinates where the norm of the
        field is zero, the norm of that vector field can be passed
        using `norm_field` argument, so that pixels at those
        coordinates are not coloured. All other parameters accepted by
        `matplotlib.pyplot.imshow` can be passed. In order to use this
        function inside Jupyter notebook `%matplotlib inline` must be
        activated after `discretisedfield` is imported.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object to which the scalar plot will be added.
        norm_field : discretisedfield.Field, optional
            A (scalar) norm field used for determining whether certain
            pixels should be coloured.

        Returns
        -------
        matplotlib.image.AxesImage object

        Raises
        ------
        ValueError
            If the field has not been sliced with a plane or its
            dimension is not 1.

        Example
        -------
        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (100, 100, 100)
        >>> n = (10, 10, 10)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, dim=1, value=2)
        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(111)
        >>> field.plane('y').imshow(ax=ax)
        <matplotlib.image.AxesImage object at ...>

        .. seealso:: :py:func:`~discretisedfield.Field.quiver`

        """
        if not hasattr(self.mesh, 'info'):
            msg = ('Only sliced field can be plotted using imshow. '
                   'For instance, field.plane(\'x\').imshow(ax=ax).')
            raise ValueError(msg)
        if self.dim > 1:
            msg = ('Only scalar (dim=1) fields can be plotted. Consider '
                   'plotting one component, e.g. field.x.imshow(ax=ax) '
                   'or norm field.norm.imshow(ax=ax).')
            raise ValueError(msg)

        points, values = list(zip(*list(self)))

        # If norm_field is passed, set values where norm=0 to np.nan,
        # so that they are not plotted.
        if norm_field is not None:
            values = list(values)  # make values mutable
            for i, point in enumerate(points):
                if norm_field(point) == 0:
                    values[i] = np.nan

            # "Unpack" values inside arrays.
            values = [v[0] if not np.isnan(v) else v for v in values]
        else:
            # "Unpack" values inside arrays.
            values = list(zip(*values))

        points = list(zip(*points))

        extent = [self.mesh.pmin[self.mesh.info['axis1']],
                  self.mesh.pmax[self.mesh.info['axis1']],
                  self.mesh.pmin[self.mesh.info['axis2']],
                  self.mesh.pmax[self.mesh.info['axis2']]]
        n = (self.mesh.n[self.mesh.info['axis2']],
             self.mesh.n[self.mesh.info['axis1']])

        imax = ax.imshow(np.array(values).reshape(n), origin='lower',
                         extent=extent, **kwargs)

        return imax

    def quiver(self, ax=None, color_field=None, **kwargs):
        """Plots a vector field plane using `matplotlib.pyplot.quiver`.

        Before the field can be plotted, it must be sliced with a
        plane (e.g. `field.plane(`y`)`) and field must be of dimension
        3 (vector field). Otherwise, ValueError is raised. `quiver`
        adds the plot to `matplotlib.axes.Axes` passed via `ax`
        argument. If there are coordinates where the norm of the field
        is zero, vectors are not plotted at those coordinates. By
        default, plot is not coloured, but by passing a
        `discretisedfield.Field` object of dimension 1 as
        `color_field`, quiver plot will be coloured based on the
        values from the field. All other parameters accepted by
        `matplotlib.pyplot.quiver` can be passed. In order to use this
        function inside Jupyter notebook `%matplotlib inline` must be
        activated after `discretisedfield` is imported.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object to which the quiver plot will be added.
        color_field : discretisedfield.Field, optional
            A (scalar) field used for determining the colour of the
            quiver plot.

        Returns
        -------
        matplotlib.quiver.Quiver object

        Raises
        ------
        ValueError
            If the field has not been sliced with a plane or its
            dimension is not 3.

        Example
        -------
        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (100, 100, 100)
        >>> n = (10, 10, 10)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, dim=3, value=(1, 2, 0))
        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(111)
        >>> field.plane(z=50).quiver(ax=ax, color_field=field.z)
        <matplotlib.quiver.Quiver object at ...>

        .. seealso:: :py:func:`~discretisedfield.Field.imshow`

        """
        if not hasattr(self.mesh, 'info'):
            msg = ('Only sliced field can be plotted using quiver. '
                   'For instance, field.plane(\'x\').quiver(ax=ax).')
            raise ValueError(msg)
        if self.dim != 3:
            msg = 'Only three-dimensional (dim=3) fields can be plotted.'
            raise ValueError(msg)

        points, values = list(zip(*list(self)))

        # Remove values where norm is 0
        points, values = list(points), list(values)  # make them mutable
        points = [p for p, v in zip(points, values)
                  if not np.equal(v, 0).all()]
        values = [v for v in values if not np.equal(v, 0).all()]
        if color_field is not None:
            colors = [color_field(p) for p in points]
            colors = list(zip(*colors))

        # "Unpack" values inside arrays.
        points, values = list(zip(*points)), list(zip(*values))

        # Are there any vectors pointing out-of-plane? If yes, set the scale.
        if not any(values[self.mesh.info['axis1']] +
                   values[self.mesh.info['axis2']]):
            kwargs['scale'] = 1

        kwargs['pivot'] = 'mid'  # arrow at the centre of the cell

        if color_field is None:
            # quiver plot is not coloured.
            qvax = ax.quiver(points[self.mesh.info['axis1']],
                             points[self.mesh.info['axis2']],
                             values[self.mesh.info['axis1']],
                             values[self.mesh.info['axis2']],
                             **kwargs)

        else:
            # quiver plot is coloured.
            qvax = ax.quiver(points[self.mesh.info['axis1']],
                             points[self.mesh.info['axis2']],
                             values[self.mesh.info['axis1']],
                             values[self.mesh.info['axis2']],
                             colors,
                             **kwargs)

        return qvax

    def colorbar(self, ax, coloredplot, cax=None, **kwargs):
        """Adds a colorbar to the axes using `matplotlib.pyplot.colorbar`.

        Axes to which the colorbar should be added is passed via `ax`
        argument. If the colorbar axes are made before the method is
        called, they should be passed as `cax`. The plot to which the
        colorbar should correspond to is passed via `coloredplot`. All
        other parameters accepted by `matplotlib.pyplot.colorbar` can
        be passed. In order to use this function inside Jupyter
        notebook `%matplotlib inline` must be activated after
        `discretisedfield` is imported.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object to which the colorbar will be added.
        coloredplot : matplotlib.quiver.Quiver, matplotlib.image.AxesImage
            A plot to which the colorbar should correspond
        cax : matplotlib.axes.Axes, optional
            Colorbar axes.

        Returns
        -------
        matplotlib.colorbar.Colorbar

        Example
        -------
        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (100, 100, 100)
        >>> n = (10, 10, 10)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, dim=3, value=(1, 2, 0))
        >>> fig = plt.figure()
        >>> ax = fig.add_subplot(111)
        >>> coloredplot = field.plane(z=50).quiver(ax=ax, color_field=field.z)
        >>> field.colorbar(ax=ax, coloredplot=coloredplot)
        <matplotlib.colorbar.Colorbar object at ...>

        """
        if cax is None:
            divider = mpl_toolkits.axes_grid1.make_axes_locatable(ax)
            cax = divider.append_axes('right', size='5%', pad=0.1)

        cbar = plt.colorbar(coloredplot, cax=cax, **kwargs)

        return cbar

    def k3d_nonzero(self, color=dfu.colormap[0], plot=None, **kwargs):
        """Plots the voxels where the value of a scalar field is nonzero.

        All mesh cells where the value of the field is not zero will
        be marked using the same color. Only scalar fields can be
        plotted. Otherwise, ValueError is raised. Different colour of
        voxels can be passed in the RGB format using `color`
        parameter. This function is often used to look at the defined
        sample in the finite difference mesh, by inspecting its norm
        (`field.norm.k3d_nonzero`). If `plot` is passed as a
        `k3d.plot.Plot`, plot is added to it. Otherwise, a new k3d
        plot is created. All arguments allowed in `k3d.voxels()` can
        be passed. This function is to be called in Jupyter notebook.

        Parameters
        ----------
        color : int/hex, optional
            Voxel color in hexadecimal format.
        plot : k3d.plot.Plot, optional
            If this argument is passed, plot is added to
            it. Otherwise, a new k3d plot is created.

        Example
        -------
        >>> import discretisedfield as df
        ...
        >>> p1 = (-50, -50, -50)
        >>> p2 = (50, 50, 50)
        >>> n = (10, 10, 10)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, dim=3, value=(1, 2, 0))
        >>> def normfun(pos):
        ...     x, y, z = pos
        ...     if x**2 + y**2 < 30**2:
        ...         return 1
        ...     else:
        ...         return 0
        >>> field.norm = normfun
        >>> field.norm.k3d_nonzero()
        Plot(...)

        .. seealso:: :py:func:`~discretisedfield.Field.k3d_voxels`
        """
        if self.dim > 1:
            msg = ('Only scalar (dim=1) fields can be plotted. Consider '
                   'plotting one component, e.g. field.x.k3d_nonzero() '
                   'or norm field.norm.k3d_nonzero().')
            raise ValueError(msg)
        plot_array = np.copy(self.array)  # make a deep copy
        plot_array = np.squeeze(plot_array)  # remove an empty dimension
        plot_array = np.swapaxes(plot_array, 0, 2)  # k3d: arrays are (z, y, x)
        plot_array[plot_array != 0] = 1  # all cells have the same colour

        # In the case of nano-sized samples, fix the order of
        # magnitude of the plot extent to avoid freezing the k3d plot.
        if np.any(np.divide(self.mesh.cell, 1e-9) < 1e3):
            pmin = np.divide(self.mesh.pmin, 1e-9)
            pmax = np.divide(self.mesh.pmax, 1e-9)
        else:
            pmin = self.mesh.pmin
            pmax = self.mesh.pmax

        dfu.voxels(plot_array, pmin, pmax, colormap=color,
                   plot=plot, **kwargs)

    def k3d_voxels(self, norm_field=None, plot=None, **kwargs):
        """Plots the scalar field as a coloured `k3d.voxels()` plot.

        At all mesh cells, a voxel will be plotted anc coloured
        according to its value. If the scalar field plotted is
        extracted from a vector field, which has coordinates where the
        norm of the field is zero, the norm of that vector field can
        be passed using `norm_field` argument, so that voxels at those
        coordinates are not showed. Only scalar fields can be
        plotted. Otherwise, ValueError is raised. If `plot` is passed
        as a `k3d.plot.Plot`, plot is added to it. Otherwise, a new
        k3d plot is created. All arguments allowed in `k3d.voxels()`
        can be passed. This function is to be called in Jupyter
        notebook.

        Parameters
        ----------
        norm_field : discretisedfield.Field, optional
            A (scalar) norm field used for determining whether certain
            voxels should be plotted.
        plot : k3d.plot.Plot, optional
            If this argument is passed, plot is added to
            it. Otherwise, a new k3d plot is created.

        Example
        -------
        >>> import discretisedfield as df
        ...
        >>> p1 = (-50, -50, -50)
        >>> p2 = (50, 50, 50)
        >>> n = (10, 10, 10)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, dim=3, value=(1, 2, 0))
        >>> def normfun(pos):
        ...     x, y, z = pos
        ...     if x**2 + y**2 < 30**2:
        ...         return 1
        ...     else:
        ...         return 0
        >>> field.norm = normfun
        >>> field.x.k3d_voxels(norm_field=field.norm)
        Plot(...)

        .. seealso:: :py:func:`~discretisedfield.Field.k3d_vectors`

        """
        if self.dim > 1:
            msg = ('Only scalar (dim=1) fields can be plotted. Consider '
                   'plotting one component, e.g. field.x.k3d_nonzero() '
                   'or norm field.norm.k3d_nonzero().')
            raise ValueError(msg)

        plot_array = np.copy(self.array)  # make a deep copy
        plot_array = plot_array[..., 0]  # remove an empty dimension

        plot_array -= plot_array.min()
        # In the case of uniform fields, division by zero can be
        # encountered.
        if plot_array.max() != 0:
            plot_array /= plot_array.max()
        plot_array *= 254
        plot_array += 1
        plot_array = plot_array.round()
        plot_array = plot_array.astype(int)

        if norm_field is not None:
            for index in self.mesh.indices:
                if norm_field(self.mesh.index2point(index)) == 0:
                    plot_array[index] = 0

        plot_array = np.swapaxes(plot_array, 0, 2)  # k3d: arrays are (z, y, x)

        cmap = matplotlib.cm.get_cmap('viridis', 256)
        colormap = [dfu.num2hexcolor(i, cmap) for i in range(cmap.N)]

        # In the case of nano-sized samples, fix the order of
        # magnitude of the plot extent to avoid freezing the k3d plot.
        if np.any(np.divide(self.mesh.cell, 1e-9) < 1e3):
            pmin = np.divide(self.mesh.pmin, 1e-9)
            pmax = np.divide(self.mesh.pmax, 1e-9)
        else:
            pmin = self.mesh.pmin
            pmax = self.mesh.pmax

        dfu.voxels(plot_array, pmin, pmax, colormap=colormap,
                   plot=plot, **kwargs)

    def k3d_vectors(self, color_field=None, points=True, plot=None, **kwargs):
        """Plots the vector field as a `k3d.vectors()` plot.

        At all mesh cells, a vector will be plotted if its norm is not
        zero. Vectors can be coloured according to the values of the
        scalar field passed as `color_field`. Only vector fields can
        be plotted. Otherwise, ValueError is raised. Points at the
        discretisation cell centres can be added by setting
        `points=True`. If `plot` is passed as a `k3d.plot.Plot`, plot
        is added to it. Otherwise, a new k3d plot is created. All
        arguments allowed in `k3d.vectors()` can be passed. This
        function is to be called in Jupyter notebook.

        Parameters
        ----------
        color_field : discretisedfield.Field, optional
            A (scalar) field used for determining the colours of
            vectors.
        points : bool, optional
            If `True`, points will be added to the discretisation cell
            centres.
        plot : k3d.plot.Plot, optional
            If this argument is passed, plot is added to
            it. Otherwise, a new k3d plot is created.

        Example
        -------
        1. Plotting an entire vector field.

        >>> import discretisedfield as df
        ...
        >>> p1 = (-50, -50, -50)
        >>> p2 = (50, 50, 50)
        >>> n = (10, 10, 10)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, dim=3, value=(1, 2, 0))
        >>> field.k3d_vectors(color_field=field.x)
        Plot(...)

        2. Plotting the slice of a vector field.

        >>> import discretisedfield as df
        ...
        >>> p1 = (-50, -50, -50)
        >>> p2 = (50, 50, 50)
        >>> n = (10, 10, 10)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> field = df.Field(mesh, dim=3, value=(1, 2, 0))
        >>> field.plane('x').k3d_vectors(color_field=field.x)
        Plot(...)

        .. seealso:: :py:func:`~discretisedfield.Field.k3d_voxels`

        """
        if self.dim != 3:
            msg = 'Only three-dimensional (dim=3) fields can be plotted.'
            raise ValueError(msg)

        coordinates, vectors, color_values = [], [], []
        norm = self.norm  # assigned to be computed only once
        for coord, value in self:
            if norm(coord) > 0:
                coordinates.append(coord)
                vectors.append(value)
                if color_field is not None:
                    color_values.append(color_field(coord)[0])

        coordinates, vectors = np.array(coordinates), np.array(vectors)

        # In the case of nano-sized samples, fix the order of
        # magnitude of the coordinates to avoid freezing the k3d plot.
        if np.any(np.divide(self.mesh.cell, 1e-9) < 1e3):
            coordinates /= 1e-9
            cell = np.divide(self.mesh.cell, 1e-9)
        else:
            cell = self.mesh.cell

        # Scale the vectors to correspond to the size of cells.
        vectors /= vectors.max()
        vectors *= 0.8*np.array(cell)

        # Middle of the arrow is at the cell centre.
        coordinates -= 0.5 * vectors

        if color_field is not None:
            color_values = np.array(color_values)
            color_values -= color_values.min()
            # In the case of uniform fields, division by zero can be
            # encountered.
            if color_values.max() != 0:
                color_values /= color_values.max()
            color_values *= 256
            color_values = color_values.round()
            color_values = color_values.astype(int)

            cmap = matplotlib.cm.get_cmap('viridis', 256)
            colors = []
            for c in color_values:
                color = dfu.num2hexcolor(c, cmap)
                colors.append((color, color))
        else:
            colors = []

        plot = dfu.vectors(coordinates, vectors, colors=colors,
                           plot=plot, **kwargs)

        if points:
            dfu.points(coordinates + 0.5 * vectors, plot=plot)



import os
import sys
import subprocess
import discretisedfield as df
from discretisedfield.ovf2vtk import convert_files


def check_vtk(vtkfile):
    pattern = ''
    with open(vtkfile, 'r') as f:
        for line in f.readlines():
            if 'CELL_DATA' in line:
                pattern = line.strip()
                break
    assert pattern == 'CELL_DATA 8'


def test_ovf2vtk():
    mesh = df.Mesh(p1=(0, 0, 0), p2=(10, 10, 10), cell=(5, 5, 5))
    f = df.Field(mesh, dim=3, value=(1, 2, -3.14))

    filename_omf = 'test_ovf2vtk_file.omf'
    filename_vtk = 'test_ovf2vtk_file.vtk'
    f.write(filename_omf)

    interpreter = sys.executable.split('/')[-1]

    # No output filename provided.
    cmd = [interpreter, '-m', 'discretisedfield.ovf2vtk',
           '--infile', f'{filename_omf}']
    proc_return = subprocess.run(cmd)
    assert proc_return.returncode == 0
    check_vtk(filename_vtk)

    # Output filename provided.
    cmd = [interpreter, '-m', 'discretisedfield.ovf2vtk',
           '--infile', f'{filename_omf}', '--outfile',
           f'{filename_vtk}']
    proc_return = subprocess.run(cmd)
    assert proc_return.returncode == 0
    check_vtk(filename_vtk)

    # Number of input and output files do not match.
    cmd = [interpreter, '-m', 'discretisedfield.ovf2vtk',
           '--infile', f'{filename_omf}', '--outfile',
           f'{filename_vtk}', f'anotherfile.vtk']
    proc_return = subprocess.run(cmd)
    assert proc_return.returncode == 0

    os.remove(f'{filename_omf}')
    os.remove(f'{filename_vtk}')



import numpy as np
import ubermagutil.typesystem as ts
import discretisedfield.util as dfu


@ts.typesystem(pmin=ts.Vector(size=3, const=True),
               pmax=ts.Vector(size=3, const=True))
class Region:
    """A rectangular region.

    A rectangular region spans between two corner points
    :math:`\\mathbf{p}_{1}` and :math:`\\mathbf{p}_{2}`.

    Parameters
    ----------
    p1, p2 : (3,) array_like
        Points between which the region spans :math:`\\mathbf{p}
        = (p_{x}, p_{y}, p_{z})`.

    Raises
    ------
    ValueError
        If the length of one or more region edges is zero.

    Examples
    --------
    1. Defining a nano-sized region

    >>> import discretisedfield as df
    ...
    >>> p1 = (-50e-9, -25e-9, 0)
    >>> p2 = (50e-9, 25e-9, 5e-9)
    >>> region = df.Region(p1=p1, p2=p2)

    2. An attempt to define a region, where one of the edge lengths is
    zero.

    >>> import discretisedfield as df
    ...
    >>> p1 = (-25, 3, 1)
    >>> p2 = (25, 6, 1)
    >>> region = df.Region(p1=p1, p2=p2)
    Traceback (most recent call last):
        ...
    ValueError: ...

    """
    def __init__(self, p1, p2):
        self.pmin = dfu.array2tuple(np.minimum(p1, p2))
        self.pmax = dfu.array2tuple(np.maximum(p1, p2))

        # Is any edge length of the region equal to zero?
        if np.equal(np.abs(np.subtract(self.pmax, self.pmin)), 0).any():
            msg = 'The length of one of the region edges is zero.'
            raise ValueError(msg)

    def __repr__(self):
        """Region representation.

        This method returns the string that can be copied in another
        Python script so that exactly the same region object could be
        defined.

        Returns
        -------
        str
           Region representation string.

        Example
        -------
        1. Getting region representation string.x

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> region = df.Region(p1=p1, p2=p2)
        >>> repr(region)
        'Region(p1=(0, 0, 0), p2=(2, 2, 1))'

        """
        return f'Region(p1={self.pmin}, p2={self.pmax})'

    def __contains__(self, item):
        """Determine whether `point` is inside the region. If it is, it returns
        `True`, otherwise `False`.

        Parameters
        ----------
        item : (3,) array_like
            The point coordinate :math:`(p_{x}, p_{y}, p_{z})`.

        Returns
        -------
        True
            If `item` is inside the region.
        False
            If `item` is outside the region.

        Example
        -------
        1. Check whether point is inside the region.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> region = df.Region(p1=p1, p2=p2)
        >>> point1 = (1, 1, 1)
        >>> point1 in region
        True
        >>> point2 = (1, 3, 1)
        >>> point2 in region
        False

        """
        if np.logical_or(np.less(item, self.pmin),
                         np.greater(item, self.pmax)).any():
            return False
        else:
            return True



import pytest
import numpy as np
import discretisedfield.util as dfu


def test_array2tuple():
    dfu.array2tuple(np.array([1, 2, 3])) == (1, 2, 3)


def test_plane_info():
    info = dfu.plane_info('x')
    assert info['planeaxis'] == 0
    assert info['axis1'] == 1
    assert info['axis2'] == 2
    assert info['point'] is None

    info = dfu.plane_info('y')
    assert info['planeaxis'] == 1
    assert info['axis1'] == 0
    assert info['axis2'] == 2
    assert info['point'] is None

    info = dfu.plane_info('z')
    assert info['planeaxis'] == 2
    assert info['axis1'] == 0
    assert info['axis2'] == 1
    assert info['point'] is None

    info = dfu.plane_info(x=0)
    assert info['planeaxis'] == 0
    assert info['axis1'] == 1
    assert info['axis2'] == 2
    assert info['point'] == 0

    info = dfu.plane_info(y=0)
    assert info['planeaxis'] == 1
    assert info['axis1'] == 0
    assert info['axis2'] == 2
    assert info['point'] == 0

    info = dfu.plane_info(z=0)
    assert info['planeaxis'] == 2
    assert info['axis1'] == 0
    assert info['axis2'] == 1
    assert info['point'] == 0

    info = dfu.plane_info(x=5)
    assert info['planeaxis'] == 0
    assert info['axis1'] == 1
    assert info['axis2'] == 2
    assert info['point'] == 5

    with pytest.raises(ValueError):
        info = dfu.plane_info('xy')
    with pytest.raises(ValueError):
        info = dfu.plane_info('zy')
    with pytest.raises(ValueError):
        info = dfu.plane_info('y', 'x')
    with pytest.raises(ValueError):
        info = dfu.plane_info('xzy')
    with pytest.raises(ValueError):
        info = dfu.plane_info('z', x=3)
    with pytest.raises(ValueError):
        info = dfu.plane_info('y', y=5)
    with pytest.raises(ValueError):
        info = dfu.plane_info(x=0, z=3)



import pytest
import pkg_resources
import matplotlib
matplotlib.use('Agg')
from .mesh import Mesh
from .field import Field
from .region import Region

__version__ = pkg_resources.get_distribution(__name__).version
__dependencies__ = pkg_resources.require(__name__)


def test():
    return pytest.main(['-v', '--pyargs',
                        'discretisedfield', '-l'])  # pragma: no cover



import types
import pytest
import numpy as np
import discretisedfield as df


def check_mesh(mesh):
    assert isinstance(mesh.p1, tuple)
    assert len(mesh.p1) == 3
    assert isinstance(mesh.p2, tuple)
    assert len(mesh.p2) == 3
    assert isinstance(mesh.cell, tuple)
    assert len(mesh.cell) == 3
    assert isinstance(mesh.n, tuple)
    assert len(mesh.n) == 3
    assert all(isinstance(i, int) for i in mesh.n)
    assert isinstance(mesh.pmin, tuple)
    assert len(mesh.pmin) == 3
    assert isinstance(mesh.pmax, tuple)
    assert len(mesh.pmax) == 3
    assert isinstance(mesh.l, tuple)
    assert len(mesh.l) == 3
    assert isinstance(mesh.ntotal, int)
    assert isinstance(mesh.pbc, set)
    assert all(isinstance(i, str) for i in mesh.pbc)
    assert isinstance(mesh.regions, dict)
    assert isinstance(mesh.name, str)


class TestMesh:
    def setup(self):
        self.valid_args = [[(0, 0, 0), (5, 5, 5),
                            [1, 1, 1], None],
                           [(-1, 0, -3), (5, 7, 5),
                            None, (1, 1, 1)],
                           [(0, 0, 0), (5e-9, 5e-9, 5e-9),
                            None, (1e-9, 1e-9, 1e-9)],
                           [(0, 0, 0), (5e-9, 5e-9, 5e-9),
                            (5, 5, 5), None],
                           [(-1.5e-9, -5e-9, 0), (1.5e-9, -15e-9, -10e-9),
                            None, (1.5e-9, 0.5e-9, 10e-9)],
                           [(-1.5e-9, -5e-9, 0), (1.5e-9, -15e-9, -10e-9),
                            (3, 50, 2), None],
                           [(-1.5e-9, -5e-9, -5e-9), np.array((0, 0, 0)),
                            None, (0.5e-9, 1e-9, 5e-9)],
                           [(-1.5e-9, -5e-9, -5e-9), np.array((0, 0, 0)),
                            (5, 5, 7), None],
                           [[0, 5e-6, 0], (-1.5e-6, -5e-6, -5e-6),
                            None, (0.5e-6, 1e-6, 2.5e-6)],
                           [[0, 5e-6, 0], (-1.5e-6, -5e-6, -5e-6),
                            (1, 10, 100), None],
                           [(0, 125e-9, 0), (500e-9, 0, -3e-9),
                            None, (2.5e-9, 2.5e-9, 3e-9)]]

        self.invalid_args = [[(0, 0, 0), (5, 5, 5),
                              None, (-1, 1, 1)],
                             [(0, 0, 0), (5, 5, 5),
                              (-1, 1, 1), None],
                             [(0, 0, 0), (5, 5, 5),
                              'n', None],
                             [(0, 0, 0), (5, 5, 5),
                              (1, 2, 2+1j), None],
                             [(0, 0, 0), (5, 5, 5),
                              (1, 2, '2'), None],
                             [('1', 0, 0), (1, 1, 1),
                              None, (0, 0, 1e-9)],
                             [(-1.5e-9, -5e-9, 'a'), (1.5e-9, 15e-9, 16e-9),
                              None, (5, 1, -1e-9)],
                             [(-1.5e-9, -5e-9, 'a'), (1.5e-9, 15e-9, 16e-9),
                              (5, 1, -1), None],
                             [(-1.5e-9, -5e-9, 0), (1.5e-9, 16e-9),
                              None, (0.1e-9, 0.1e-9, 1e-9)],
                             [(-1.5e-9, -5e-9, 0), (1.5e-9, 15e-9, 1+2j),
                              None, (5, 1, 1e-9)],
                             ['string', (5, 1, 1e-9),
                              None, 'string'],
                             [(-1.5e-9, -5e-9, 0), (1.5e-9, 15e-9, 16e-9),
                              None, 2+2j]]

    def test_init_with_cell(self):
        p1 = (0, -4, 16.5)
        p2 = (15, 10.1, 11)
        cell = (1, 0.1, 0.5)
        pbc = 'yxz'
        name = 'test_mesh'
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell, pbc=pbc, name=name)

        check_mesh(mesh)
        assert mesh.p1 == p1
        assert mesh.p2 == p2
        assert mesh.cell == cell
        assert mesh.pbc == set(pbc)
        assert mesh.name == name
        assert mesh.pmin == (0, -4, 11)
        assert mesh.pmax == (15, 10.1, 16.5)
        assert mesh.l[0] == 15 - 0
        assert mesh.l[1] == 10.1 - (-4)
        assert mesh.l[2] == 16.5 - 11
        assert mesh.n[0] == (15 - 0) / 1.0
        assert mesh.n[1] == (10.1 - (-4)) / 0.1
        assert mesh.n[2] == (16.5 - 11) / 0.5
        assert mesh.ntotal == 23265

    def test_init_with_n(self):
        p1 = (0, -4, 16.5)
        p2 = (15, 10.1, 11)
        n = (15, 141, 11)
        pbc = 'yx'
        name = 'test_mesh'
        mesh = df.Mesh(p1=p1, p2=p2, n=n, pbc=pbc, name=name)

        check_mesh(mesh)
        assert mesh.p1 == p1
        assert mesh.p2 == p2
        assert mesh.pbc == set(pbc)
        assert mesh.name == name
        assert mesh.pmin == (0, -4, 11)
        assert mesh.pmax == (15, 10.1, 16.5)
        assert mesh.l[0] == 15 - 0
        assert mesh.l[1] == 10.1 - (-4)
        assert mesh.l[2] == 16.5 - 11
        assert mesh.cell[0] == (15 - 0) / 15
        assert mesh.cell[1] == (10.1 - (-4)) / 141
        assert mesh.cell[2] == (16.5 - 11) / 11
        assert mesh.ntotal == 15*141*11

    def test_init_with_n_and_cell(self):
        p1 = (0, -4, 16.5)
        p2 = (15, 10.1, 11)
        n = (15, 141, 11)
        cell = (1, 0.1, 0.5)
        pbc = 'x'
        name = 'test_mesh'
        with pytest.raises(ValueError) as excinfo:
            mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell, pbc=pbc, name=name)
        assert 'One and only one' in str(excinfo.value)

    def test_init_valid_args(self):
        for p1, p2, n, cell in self.valid_args:
            mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell)

            check_mesh(mesh)
            assert mesh.name == 'mesh'  # default name value
            assert mesh.pbc == set()  # default pbc value

    def test_init_invalid_args(self):
        for p1, p2, n, cell in self.invalid_args:
            with pytest.raises((TypeError, ValueError)):
                # Exceptions are raised by descriptors.
                mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell)

    def test_init_invalid_name(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        n = (10, 100, 2)
        for name in ['mesh name', '2name', ' ', 5, '-name', '+mn']:
            with pytest.raises((TypeError, ValueError)):
                # Exception is raised by the descriptor for mesh.name.
                mesh = df.Mesh(p1=p1, p2=p2, n=n, name=name)

    def test_init_valid_pbc(self):
        for p1, p2, n, cell in self.valid_args:
            for pbc in ['x', 'z', 'zx', 'yxzz', 'yz']:
                mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell, pbc=pbc)

                check_mesh(mesh)
                assert mesh.pbc == set(pbc)

    def test_init_invalid_pbc(self):
        for p1, p2, n, cell in self.valid_args:
            for pbc in ['abc', 'a', '123', 5]:
                with pytest.raises((ValueError, TypeError)):
                    # Exception is raised by the descriptor for mesh.pbc.
                    mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell, pbc=pbc)

    def test_zero_domain_edge_length(self):
        args = [[(0, 100e-9, 1e-9), (150e-9, 100e-9, 6e-9),
                 None, (1e-9, 0.01e-9, 1e-9)],
                [(0, 100e-9, 0), (150e-9, 101e-9, 0),
                 None, (2e-9, 1e-9, 0.1e-9)],
                [(0, 101e-9, -1), (150e-9, 101e-9, 0),
                 (1, 1, 1), None],
                [(10e9, 10e3, 0), (0.01e12, 11e3, 5),
                 (5, 100, 30), None]]

        for p1, p2, n, cell in args:
            with pytest.raises(ValueError) as excinfo:
                mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell)
            assert 'is zero' in str(excinfo.value)

    def test_domain_not_aggregate_of_cell(self):
        args = [[(0, 100e-9, 1e-9), (150e-9, 120e-9, 6e-9),
                 None, (4e-9, 1e-9, 1e-9)],
                [(0, 100e-9, 0), (150e-9, 104e-9, 1e-9),
                 None, (2e-9, 1.5e-9, 0.1e-9)],
                [(10e9, 10e3, 0), (11e9, 11e3, 5),
                 None, (1e9, 1e3, 1.5)]]

        for p1, p2, n, cell in args:
            with pytest.raises(ValueError) as excinfo:
                mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell)
            assert 'not an aggregate' in str(excinfo.value)

    def test_cell_greater_than_domain(self):
        p1 = (0, 0, 0)
        p2 = (1e-9, 1e-9, 1e-9)
        args = [(2e-9, 1e-9, 1e-9), (1e-9, 2e-9, 1e-9),
                (1e-9, 1e-9, 2e-9), (1e-9, 5e-9, 0.1e-9)]
        for cell in args:
            with pytest.raises(ValueError) as excinfo:
                mymesh = df.Mesh(p1=p1, p2=p2, cell=cell)
            assert 'not an aggregate' in str(excinfo.value)

    def test_regions(self):
        p1 = (0, 0, 0)
        p2 = (100, 50, 10)
        cell = (1, 1, 1)
        regions = {'r1': df.Region(p1=(0, 0, 0), p2=(50, 50, 10)),
                   'r2': df.Region(p1=(50, 0, 0), p2=(100, 50, 10))}
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell, regions=regions)

        assert (0, 0, 0) in mesh.regions['r1']
        assert (0, 0, 0) not in mesh.regions['r2']
        assert (25, 25, 5) in mesh.regions['r1']
        assert (25, 25, 5) not in mesh.regions['r2']
        assert (51, 10, 10) in mesh.regions['r2']
        assert (51, 10, 10) not in mesh.regions['r1']
        assert (100, 50, 10) in mesh.regions['r2']
        assert (100, 50, 10) not in mesh.regions['r1']

    def test_centre(self):
        p1 = (0, 0, 0)
        p2 = (100, 100, 100)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        assert mesh.centre == (50, 50, 50)
        assert mesh.centre in mesh

        p1 = (-18.5, 10, 0)
        p2 = (10, 5, 10)
        cell = (0.1, 0.25, 2)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        assert mesh.centre == (-4.25, 7.5, 5)
        assert mesh.centre in mesh

        p1 = (500e-9, 125e-9, 3e-9)
        p2 = (0, 0, 0)
        n = (250, 125, 3)
        mesh = df.Mesh(p1=p1, p2=p2, n=n)
        assert mesh.centre == (250e-9, 62.5e-9, 1.5e-9)
        assert mesh.centre in mesh

    def test_random_point(self):
        p1 = (-18.5, 5, 0)
        p2 = (10, -10, 10e-9)
        n = (100, 5, 10)
        mesh = df.Mesh(p1=p1, p2=p2, n=n)

        for _ in range(50):
            random_point = mesh.random_point()
            assert isinstance(random_point, tuple)
            assert random_point in mesh

    def test_repr_no_pbc(self):
        p1 = (-1, -4, 11)
        p2 = (15, 10.1, 12.5)
        cell = (1, 0.1, 0.5)
        name = 'meshname'
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell, name=name)

        rstr = ('Mesh(p1=(-1, -4, 11), p2=(15, 10.1, 12.5), '
                'cell=(1, 0.1, 0.5), pbc=set(), name=\'meshname\')')
        assert repr(mesh) == rstr

    def test_repr_pbc(self):
        p1 = (-1, -4, 11)
        p2 = (15, 10, 12.5)
        n = (16, 140, 3)
        pbc = 'x'
        name = 'meshname'
        mesh = df.Mesh(p1=p1, p2=p2, n=n, pbc=pbc, name=name)

        rstr = ('Mesh(p1=(-1, -4, 11), p2=(15, 10, 12.5), '
                'cell=(1.0, 0.1, 0.5), pbc={\'x\'}, name=\'meshname\')')
        assert repr(mesh) == rstr

    def test_index2point(self):
        p1 = (15, -4, 12.5)
        p2 = (-1, 10.1, 11)
        cell = (1, 0.1, 0.5)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

        assert mesh.index2point((5, 10, 1)) == (4.5, -2.95, 11.75)

        # Correct minimum index
        assert isinstance(mesh.index2point((0, 0, 0)), tuple)
        assert mesh.index2point((0, 0, 0)) == (-0.5, -3.95, 11.25)
        # Below minimum index
        with pytest.raises(ValueError):
            mesh.index2point((-1, 0, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, -1, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, 0, -1))

        # Correct maximum index
        assert isinstance(mesh.index2point((15, 140, 2)), tuple)
        assert mesh.index2point((15, 140, 2)) == (14.5, 10.05, 12.25)
        # Above maximum index
        with pytest.raises(ValueError):
            mesh.index2point((16, 0, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, 141, 0))
        with pytest.raises(ValueError):
            mesh.index2point((0, 0, 3))

    def test_point2index(self):
        p1 = (-10e-9, -5e-9, 10e-9)
        p2 = (10e-9, 5e-9, 0)
        cell = (1e-9, 5e-9, 1e-9)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

        # (0, 0, 0) cell
        assert mesh.point2index((-10e-9, -5e-9, 0)) == (0, 0, 0)
        assert mesh.point2index((-9.5e-9, -2.5e-9, 0.5e-9)) == (0, 0, 0)
        assert mesh.point2index((-9.01e-9, -0.1e-9, 0.9e-9)) == (0, 0, 0)

        # (19, 1, 9) cell
        assert mesh.point2index((10e-9, 5e-9, 10e-9)) == (19, 1, 9)
        assert mesh.point2index((9.5e-9, 2.5e-9, 9.5e-9)) == (19, 1, 9)
        assert mesh.point2index((9.1e-9, 0.1e-9, 9.1e-9)) == (19, 1, 9)

        # vicinity of (0, 0, 0) point
        assert mesh.point2index((1e-16, 1e-16, 0.99e-16)) == (10, 1, 0)
        assert mesh.point2index((-1e-16, -1e-16, 0.01e-16)) == (9, 0, 0)

        p1 = (-10, 5, 0)
        p2 = (10, -5, 10e-9)
        n = (10, 5, 5)
        mesh = df.Mesh(p1=p1, p2=p2, n=n)

        tol = 1e-12  # picometer tolerance
        with pytest.raises(ValueError):
            mesh.point2index((-10-tol, 0, 5))
        with pytest.raises(ValueError):
            mesh.point2index((-5, -5-tol, 5))
        with pytest.raises(ValueError):
            mesh.point2index((-5, 0, -tol))
        with pytest.raises(ValueError):
            mesh.point2index((10+tol, 0, 5))
        with pytest.raises(ValueError):
            mesh.point2index((6, 5+tol, 5))
        with pytest.raises(ValueError):
            mesh.point2index((0, 0, 10e-9+tol))

    def test_index2point_point2index_mutually_inverse(self):
        p1 = (15, -4, 12.5)
        p2 = (-1, 10.1, 11)
        cell = (1, 0.1, 0.5)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

        for p in [(-0.5, -3.95, 11.25), (14.5, 10.05, 12.25)]:
            assert isinstance(mesh.index2point(mesh.point2index(p)), tuple)
            assert mesh.index2point(mesh.point2index(p)) == p

        for i in [(1, 0, 0), (0, 1, 0), (0, 0, 1)]:
            assert isinstance(mesh.point2index(mesh.index2point(i)), tuple)
            assert mesh.point2index(mesh.index2point(i)) == i

    def test_indices(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        n = (10, 10, 10)
        mesh = df.Mesh(p1=p1, p2=p2, n=n)

        assert len(list(mesh.indices)) == 1000
        for index in mesh.indices:
            assert isinstance(index, tuple)
            assert len(index) == 3
            assert all([0 <= i <= 9 for i in index])

    def test_coordinates(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

        assert len(list(mesh.coordinates)) == 1000
        for coord in mesh.coordinates:
            assert isinstance(coord, tuple)
            assert len(coord) == 3
            assert all([0.5 <= i <= 9.5 for i in coord])

    def test_iter(self):
        p1 = (1, 1, 1)
        p2 = (5, 5, 5)
        n = (2, 2, 2)
        mesh = df.Mesh(p1=p1, p2=p2, n=n)

        assert len(list(mesh)) == 8
        for coord in mesh:
            assert isinstance(coord, tuple)
            assert len(coord) == 3
            assert all([1.5 <= i <= 4.5 for i in coord])

    def test_contains(self):
        p1 = (0, 0, 0)
        p2 = (10e-9, 10e-9, 20e-9)
        cell = (1e-9, 1e-9, 1e-9)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

        assert (0, 0, 0) in mesh
        assert (10e-9, 10e-9, 10e-9) in mesh
        assert (5e-9, 5e-9, 5e-9) in mesh
        assert (11e-9, 11e-9, 11e-9) not in mesh
        assert (-1e-9, -1e-9, -1e-9) not in mesh

    def test_line(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

        tol = 1e-12
        line = mesh.line((0, 0, 0), (10, 10, 10), n=10)
        assert isinstance(line, types.GeneratorType)
        assert len(list(line))
        for point in line:
            assert isinstance(point, tuple)
            assert len(point) == 3
            assert all([0 <= i <= 10 for i in point])

        line = list(mesh.line((0, 0, 0), (10, 0, 0), n=30))
        assert len(line) == 30
        assert line[0] == (0, 0, 0)
        assert line[-1] == (10, 0, 0)

        with pytest.raises(ValueError):
            line = list(mesh.line((-1e-9, 0, 0), (10, 0, 0), n=30))

        with pytest.raises(ValueError):
            line = list(mesh.line((0, 0, 0), (11, 0, 0), n=30))

    def test_plane(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        cell = (1, 1, 1)
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell)

        plane = mesh.plane(z=3, n=(2, 2))
        assert isinstance(plane, df.Mesh)
        assert len(list(plane)) == 4
        for point in plane:
            assert isinstance(point, tuple)
            assert len(point) == 3
            assert point[2] == 3

        plane = mesh.plane(y=9.2, n=(3, 2))
        assert isinstance(plane, df.Mesh)
        assert len(list(plane)) == 6
        for point in plane:
            assert isinstance(point, tuple)
            assert len(point) == 3
            assert point[1] == 9.2

        plane = mesh.plane('x')
        assert isinstance(plane, df.Mesh)
        assert len(list(plane)) == 100
        for point in plane:
            assert isinstance(point, tuple)
            assert len(point) == 3
            assert point[0] == 5

        with pytest.raises(ValueError):
            plane = list(mesh.plane(x=-1))

        with pytest.raises(ValueError):
            plane = list(mesh.plane(y=11))

        with pytest.raises(ValueError):
            plane = list(mesh.plane(z=-1e-9))

        with pytest.raises(ValueError):
            plane = list(mesh.plane('x', z=-1e-9))

        with pytest.raises(ValueError):
            plane = list(mesh.plane('z', z=-1e-9))

    def test_mpl(self):
        for p1, p2, n, cell in self.valid_args:
            mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell)
            mesh.mpl()

    def test_k3d(self):
        for p1, p2, n, cell in self.valid_args:
            mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell)
            mesh.k3d()

    def test_k3d_points(self):
        for p1, p2, n, cell in self.valid_args:
            mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell)
            mesh.k3d_points()

    def test_k3d_regions(self):
        p1 = (0, 0, 0)
        p2 = (100, 80, 10)
        cell = (2, 2, 2)
        regions = {'r1': df.Region(p1=(0, 0, 0), p2=(100, 10, 10)),
                   'r2': df.Region(p1=(0, 10, 0), p2=(100, 20, 10)),
                   'r3': df.Region(p1=(0, 20, 0), p2=(100, 30, 10)),
                   'r4': df.Region(p1=(0, 30, 0), p2=(100, 40, 10)),
                   'r5': df.Region(p1=(0, 40, 0), p2=(100, 50, 10)),
                   'r6': df.Region(p1=(0, 50, 0), p2=(100, 60, 10)),
                   'r7': df.Region(p1=(0, 60, 0), p2=(100, 70, 10)),
                   'r8': df.Region(p1=(0, 70, 0), p2=(100, 80, 10))}
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell, regions=regions)
        mesh.k3d_regions()

    def test_script(self):
        for p1, p2, n, cell in self.valid_args:
            mesh = df.Mesh(p1=p1, p2=p2, n=n, cell=cell)
            with pytest.raises(NotImplementedError):
                script = mesh._script

    def test_change_const_attributes(self):
        p1 = (0, 0, 0)
        p2 = (10, 10, 10)
        cell = (1, 1, 1)
        name = 'object_name'
        mesh = df.Mesh(p1=p1, p2=p2, cell=cell, name=name)

        assert mesh.p1 == p1
        assert mesh.p2 == p2
        assert mesh.cell == cell
        assert mesh.name == name
        assert mesh.l == (10, 10, 10)
        assert mesh.n == (10, 10, 10)

        # Attempt to change attribute
        with pytest.raises(AttributeError):
            mesh.p2 = (15, 15, 15)
        with pytest.raises(AttributeError):
            mesh.name = 'new_object_name'

        # Make sure the attributes have not changed.
        assert mesh.p1 == p1
        assert mesh.p2 == p2
        assert mesh.cell == cell
        assert mesh.name == name
        assert mesh.l == (10, 10, 10)
        assert mesh.n == (10, 10, 10)



import random
import itertools
import numpy as np
import discretisedfield as df
import matplotlib.pyplot as plt
import ubermagutil.typesystem as ts
import discretisedfield.util as dfu
from mpl_toolkits.mplot3d import Axes3D


@ts.typesystem(p1=ts.Vector(size=3, const=True),
               p2=ts.Vector(size=3, const=True),
               cell=ts.Vector(size=3, positive=True, const=True),
               n=ts.Vector(size=3, component_type=int, unsigned=True,
                           const=True),
               pbc=ts.Subset(sample_set="xyz"),
               name=ts.Name(const=True))
class Mesh:
    """Finite difference mesh.

    A rectangular mesh domain spans between two points
    :math:`\\mathbf{p}_{1}` and :math:`\\mathbf{p}_{2}`. The domain is
    discretised using a finite difference cell, whose dimensions are
    defined with `cell`. Alternatively, the domain can be discretised
    by passing the number of discretisation cells `n` in all three
    dimensions. Either `cell` or `n` should be defined to discretise
    the domain, not both. Periodic boundary conditions can be
    specified by passing `pbc` argument, which is an iterable
    containing one or more elements from ``['x', 'y', 'z']``. The
    parameter `name` is optional and defaults to 'mesh'.

    In order to properly define a mesh, the length of all mesh domain
    edges must not be zero and the mesh domain must be an aggregate of
    discretisation cells.

    Parameters
    ----------
    p1, p2 : (3,) array_like
        Points between which the mesh domain spans :math:`\\mathbf{p}
        = (p_{x}, p_{y}, p_{z})`.
    cell : (3,) array_like, optional
        Discretisation cell size :math:`(d_{x}, d_{y}, d_{z})`. Either
        `cell` or `n` should be defined, not both.
    n : (3,) array_like, optional
        The number of discretisation cells :math:`(n_{x}, n_{y},
        n_{z})`. Either `cell` or `n` should be defined, not both.
    pbc : iterable, optional
        Periodic boundary conditions in x, y, or z direction. Its value
        is an iterable consisting of one or more characters `x`, `y`,
        or `z`, denoting the direction(s) in which the mesh is periodic.
    regions : dict, optional
        A dictionary defining regions inside the mesh. The keys of the
        dictionary are the region names (str), whereas the values are
        `discretisedfield.Region` objects.
    name : str, optional
        Mesh name (the default is 'mesh'). The mesh name must be a valid
        Python variable name string. More specifically, it must not
        contain spaces, or start with underscore or numeric character.

    Raises
    ------
    ValueError
        If the length of one or more mesh domain edges is zero, or
        mesh domain is not an aggregate of discretisation cells.

    Examples
    --------
    1. Defining a nano-sized thin film mesh by passing `cell` parameter

    >>> import discretisedfield as df
    ...
    >>> p1 = (-50e-9, -25e-9, 0)
    >>> p2 = (50e-9, 25e-9, 5e-9)
    >>> cell = (1e-9, 1e-9, 0.1e-9)
    >>> name = "mesh_name"
    >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell, name=name)

    2. Defining a nano-sized thin film mesh by passing `n` parameter

    >>> import discretisedfield as df
    ...
    >>> p1 = (-50e-9, -25e-9, 0)
    >>> p2 = (50e-9, 25e-9, 5e-9)
    >>> n = (100, 50, 5)
    >>> name = "mesh_name"
    >>> mesh = df.Mesh(p1=p1, p2=p2, n=n, name=name)

    3. Defining a mesh with periodic boundary conditions in x and y
    directions.

    >>> import discretisedfield as df
    ...
    >>> p1 = (0, 0, 0)
    >>> p2 = (100, 100, 1)
    >>> n = (100, 100, 1)
    >>> pbc = 'xy'
    >>> mesh = df.Mesh(p1=p1, p2=p2, n=n, pbc=pbc, name=name)

    4. Defining a mesh with two regions.

    >>> import discretisedfield as df
    ...
    >>> p1 = (0, 0, 0)
    >>> p2 = (100, 100, 100)
    >>> n = (10, 10, 10)
    >>> regions = {'r1': df.Region(p1=(0, 0, 0), p2=(50, 100, 100)),
    ...            'r2': df.Region(p1=(50, 0, 0), p2=(100, 100, 100))}
    >>> mesh = df.Mesh(p1=p1, p2=p2, n=n, regions=regions)

    5. An attempt to define a mesh with invalid parameters, so that
    the ``ValueError`` is raised. In this example, the mesh domain is
    not an aggregate of discretisation cells in the :math:`z`
    direction.

    >>> import discretisedfield as df
    ...
    >>> p1 = (-25, 3, 0)
    >>> p2 = (25, 6, 1)
    >>> cell = (5, 3, 0.4)
    >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell, name=name)
    Traceback (most recent call last):
        ...
    ValueError: ...

    """
    def __init__(self, p1, p2, n=None, cell=None, pbc=set(),
                 regions={}, name='mesh'):
        self.p1 = tuple(p1)
        self.p2 = tuple(p2)
        self.pbc = pbc
        self.name = name
        self.regions = regions

        # Is any edge length of the domain equal to zero?
        if np.equal(self.l, 0).any():
            msg = 'The length of one of the domain edges is zero.'
            raise ValueError(msg)

        # Determine whether cell or n was passed and define them both.
        if cell is not None and n is None:
            self.cell = tuple(cell)
            n = np.divide(self.l, self.cell).round().astype(int)
            self.n = dfu.array2tuple(n)
        elif n is not None and cell is None:
            self.n = tuple(n)
            cell = np.divide(self.l, self.n).astype(float)
            self.cell = dfu.array2tuple(cell)
        else:
            msg = ('One and only one of the parameters '
                   '(n or cell) should be defined.')
            raise ValueError(msg)

        # Is the mesh domain not an aggregate of discretisation cells?
        tol = 1e-12  # picometre tolerance
        rem = np.remainder(self.l, self.cell)
        if np.logical_and(np.greater(rem, tol),
                          np.less(rem, np.subtract(self.cell, tol))).any():
            msg = 'Mesh domain is not an aggregate of the discretisation cell.'
            raise ValueError(msg)

    @property
    def pmin(self):
        """Mesh point with minimum coordinates.

        The :math:`i`-th component of :math:`\\mathbf{p}_\\text{min}`
        is computed from points :math:`p_{1}` and :math:`p_{2}`
        between which the mesh domain spans: :math:`p_\\text{min}^{i}
        = \\text{min}(p_{1}^{i}, p_{2}^{i})`.

        Returns
        -------
        tuple (3,)
            Point with minimum mesh coordinates :math:`(p_{x}^\\text{min},
            p_{y}^\\text{min}, p_{z}^\\text{min})`.

        Examples
        --------
        1. Getting the minimum mesh point.

        >>> import discretisedfield as df
        ...
        >>> p1 = (-1.1, 2.9, 0)
        >>> p2 = (5, 0, -0.1)
        >>> cell = (0.1, 0.1, 0.005)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell, name='mesh')
        >>> mesh.pmin
        (-1.1, 0.0, -0.1)

        .. seealso:: :py:func:`~discretisedfield.Mesh.pmax`

        """
        res = np.minimum(self.p1, self.p2)
        return dfu.array2tuple(res)

    @property
    def pmax(self):
        """Mesh point with maximum coordinates.

        The :math:`i`-th component of :math:`\\mathbf{p}_\\text{max}`
        is computed from points :math:`p_{1}` and :math:`p_{2}`
        between which the mesh domain spans: :math:`p_\\text{min}^{i}
        = \\text{max}(p_{1}^{i}, p_{2}^{i})`.

        Returns
        -------
        tuple (3,)
            Point with maximum mesh coordinates :math:`(p_{x}^\\text{max},
            p_{y}^\\text{max}, p_{z}^\\text{max})`.

        Examples
        --------
        1. Getting the maximum mesh point.

        >>> import discretisedfield as df
        ...
        >>> p1 = (-1.1, 2.9, 0)
        >>> p2 = (5, 0, -0.1)
        >>> cell = (0.1, 0.1, 0.005)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.pmax
        (5.0, 2.9, 0.0)

        .. seealso:: :py:func:`~discretisedfield.Mesh.pmin`

        """
        res = np.maximum(self.p1, self.p2)
        return dfu.array2tuple(res)

    @property
    def l(self):
        """Mesh domain edge lengths.

        Edge length in any direction :math:`i` is computed from the
        points between which the mesh domain spans :math:`l^{i} =
        |p_{2}^{i} - p_{1}^{i}|`.

        Returns
        -------
        tuple (3,)
            Lengths of mesh domain edges :math:`(l_{x}, l_{y}, l_{z})`.

        Examples
        --------
        1. Getting the mesh domain edge lengths.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, -5)
        >>> p2 = (5, 15, 15)
        >>> n = (5, 15, 20)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> mesh.l
        (5, 15, 20)

        .. seealso:: :py:func:`~discretisedfield.Mesh.n`

        """
        res = np.abs(np.subtract(self.p1, self.p2))
        return dfu.array2tuple(res)

    @property
    def ntotal(self):
        """Total number of discretisation cells.

        It is computed by multiplying all elements of `self.n`.

        Returns
        -------
        int
            The total number of discretisation cells

        Examples
        --------
        1. Getting the number of discretisation cells in a mesh.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 5, 0)
        >>> p2 = (5, 15, 2)
        >>> cell = (1, 0.1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.ntotal
        1000

        .. seealso:: :py:func:`~discretisedfield.Mesh.n`

        """
        return int(np.prod(self.n))

    @property
    def centre(self):
        """Mesh domain centre point.

        It is computed as the middle point between minimum and maximum
        coordinates :math:`\\mathbf{p}_\\text{c} = \\frac{1}{2}
        (\\mathbf{p}_\\text{min} + \\mathbf{p}_\\text{max})` This
        point does not necessarily coincide with the discretisation
        cell centre.

        Returns
        -------
        tuple (3,)
            Mesh domain centre point :math:`(p_{c}^{x}, p_{c}^{y},
            p_{c}^{z})`.

        Examples
        --------
        1. Getting the mesh centre point.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (5, 15, 20)
        >>> n = (5, 15, 20)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> mesh.centre
        (2.5, 7.5, 10.0)

        """
        res = np.multiply(np.add(self.pmin, self.pmax), 0.5)
        return dfu.array2tuple(res)

    @property
    def indices(self):
        """Generator yielding indices of all mesh cells.

        Yields
        ------
        tuple (3,)
            Mesh cell indices :math:`(i_{x}, i_{y}, i_{z})`.

        Examples
        --------
        1. Getting all mesh cell indices.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (3, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> list(mesh.indices)
        [(0, 0, 0), (1, 0, 0), (2, 0, 0), (0, 1, 0), (1, 1, 0), (2, 1, 0)]

        .. seealso:: :py:func:`~discretisedfield.Mesh.coordinates`

        """
        for index in itertools.product(*map(range, reversed(self.n))):
            yield tuple(reversed(index))

    @property
    def coordinates(self):
        """Generator yielding coordinates of all mesh cells.

        The discretisation cell coordinate corresponds to the cell
        centre point.

        Yields
        ------
        tuple (3,)
            Mesh cell coordinates (`px`, `py`, `pz`).

        Examples
        --------
        1. Getting all mesh cell coordinates.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> list(mesh.coordinates)
        [(0.5, 0.5, 0.5), (1.5, 0.5, 0.5), (0.5, 1.5, 0.5), (1.5, 1.5, 0.5)]

        .. seealso:: :py:func:`~discretisedfield.Mesh.indices`

        """
        for index in self.indices:
            yield self.index2point(index)

    def __iter__(self):
        """This method makes `df.Mesh` object iterable.

        It iterates through the coodinates of the mesh cells
        (`df.Mesh.coordinates`).

        .. seealso:: :py:func:`~discretisedfield.Mesh.indices`

        """
        return self.coordinates

    def __contains__(self, item):
        """Determine whether `point` is inside the mesh. If it is, it returns
        `True`, otherwise `False`.

        Parameters
        ----------
        item : (3,) array_like
            The mesh point coordinate :math:`(p_{x}, p_{y}, p_{z})`.

        Returns
        -------
        True
            If `item` is inside the mesh.
        False
            If `item` is outside the mesh.

        Examples
        --------
        1. Check whether point is inside the mesh.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> point1 = (1, 1, 1)
        >>> point1 in mesh
        True
        >>> point2 = (1, 3, 1)
        >>> point2 in mesh
        False

        """
        region = df.Region(self.p1, self.p2)
        return item in region

    def __repr__(self):
        """Mesh representation string.

        This method returns the string that can be copied in another
        Python script so that exactly the same mesh object could be
        defined.

        Returns
        -------
        str
            Mesh representation string.

        Examples
        --------
        1. Getting mesh representation string.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> pbc = 'x'
        >>> name = 'm'
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell, pbc=pbc, name=name)
        >>> repr(mesh)
        "Mesh(p1=(0, 0, 0), p2=(2, 2, 1), cell=(1, 1, 1), pbc={'x'}, name='m')"

        """
        return (f'Mesh(p1={self.p1}, p2={self.p2}, cell={self.cell}, '
                f'pbc={self.pbc}, name=\'{self.name}\')')

    def random_point(self):
        """Generate the random point belonging to the mesh.

        The use of this function is mostly limited for writing tests
        for packages based on `discretisedfield`.

        Returns
        -------
        tuple (3,)
            Coordinates of a random point inside that belongs to the mesh
            :math:`(x_\\text{rand}, y_\\text{rand}, z_\\text{rand})`.

        Examples
        --------
        1. Generating a random mesh point.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (200e-9, 200e-9, 1e-9)
        >>> cell = (1e-9, 1e-9, 0.5e-9)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.random_point()
        (...)

        .. note::

           In the example, ellipsis is used instead of an exact tuple
           because the result differs each time ``random_point``
           method is called.

        """
        res = np.add(self.pmin, np.multiply(np.random.random(3), self.l))
        return dfu.array2tuple(res)

    def index2point(self, index):
        """Convert cell's index to the cell's centre coordinate.

        Parameters
        ----------
        index : (3,) array_like
            The cell's index :math:`(i_{x}, i_{y}, i_{z})`.

        Returns
        -------
            The cell's centre point :math:`(p_{x}, p_{y}, p_{z})`.

        Raises
        ------
        ValueError
            If the cell's index is out of range.

        Examples
        --------
        1. Converting cell's index to its coordinate.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.index2point((0, 1, 0))
        (0.5, 1.5, 0.5)

        .. seealso:: :py:func:`~discretisedfield.Mesh.point2index`

        """
        # Does index refer to a cell outside the mesh?
        if np.logical_or(np.less(index, 0),
                         np.greater_equal(index, self.n)).any():
            msg = 'Index out of range.'
            raise ValueError(msg)

        res = np.add(self.pmin, np.multiply(np.add(index, 0.5), self.cell))
        return dfu.array2tuple(res)

    def point2index(self, point):
        """Compute the index of a cell to which the point belongs to.

        Parameters
        ----------
        p : (3,) array_like
            The point's coordinate :math:`(p_{x}, p_{y}, p_{z})`.

        Returns
        -------
            The cell's index :math:`(i_{x}, i_{y}, i_{z})`.

        Raises
        ------
        ValueError
            If `point` is outside the mesh domain.

        Examples
        --------
        1. Converting point's coordinate to the cell index.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 1)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.point2index((0.2, 1.7, 0.3))
        (0, 1, 0)

        .. seealso:: :py:func:`~discretisedfield.Mesh.index2point`

        """
        if point not in self:
            msg = f'Point {point} is outside the mesh.'
            raise ValueError(msg)

        index = np.subtract(np.divide(np.subtract(point, self.pmin),
                                      self.cell), 0.5).round().astype(int)
        # If index is rounded to the out-of-range values.
        index = np.clip(index, 0, np.subtract(self.n, 1))

        return dfu.array2tuple(index)

    def line(self, p1, p2, n=100):
        """Line generator.

        Given two points :math:`p_{1}` and :math:`p_{2}`, :math:`n`
        position coordinates are generated.

        .. math::

           \\mathbf{r}_{i} = i\\frac{\\mathbf{p}_{2} -
           \\mathbf{p}_{1}}{n-1}

        and this method yields :math:`\\mathbf{r}_{i}` in :math:`n`
        iterations.

        Parameters
        ----------
        p1, p2 : (3,) array_like
            Two points between which the line is generated.
        n : int
            Number of points on the line.

        Yields
        ------
        tuple
            :math:`\\mathbf{r}_{i}`

        Raises
        ------
        ValueError
            If `p1` or `p2` is outside the mesh domain.

        Examples
        --------
        1. Creating a line generator.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 2)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> tuple(mesh.line(p1=(0, 0, 0), p2=(2, 0, 0), n=2))
        ((0.0, 0.0, 0.0), (2.0, 0.0, 0.0))

        """
        if p1 not in self or p2 not in self:
            msg = f'Point {p1} or point {p2} is outside the mesh.'
            raise ValueError(msg)

        dl = np.subtract(p2, p1)/(n-1)
        for i in range(n):
            yield dfu.array2tuple(np.add(p1, i*dl))

    def plane(self, *args, n=None, **kwargs):
        """Slices the mesh with a plane.

        If one of the axes (`'x'`, `'y'`, or `'z'`) is passed as a
        string, a plane perpendicular to that axis is generated which
        intersects the mesh at its centre. Alternatively, if a keyword
        argument is passed (e.g. `x=1`), a plane perpendicular to the
        x-axis and intersecting it at x=1 is generated. The number of
        points in two dimensions on the plane can be defined using `n`
        (e.g. `n=(10, 15)`). Using the generated plane, a new
        "two-dimensional" mesh is created and returned.

        Parameters
        ----------
        n : tuple of length 2
            The number of points on the plane in two dimensions

        Returns
        ------
        discretisedfield.Mesh
            A mesh obtained as an intersection of mesh and the plane.

        Examples
        --------
        1. Intersecting the mesh with a plane.

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (2, 2, 2)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.plane(y=1)
        Mesh(p1=(0.0, 0.5, 0.0), p2=(2.0, 1.5, 2.0), ...)

        """
        info = dfu.plane_info(*args, **kwargs)

        if info['point'] is None:
            # Plane info was extracted from args.
            info['point'] = self.centre[info['planeaxis']]
        else:
            # Plane info was extracted from kwargs and should be
            # tested whether it is inside the mesh.
            test_point = list(self.centre)
            test_point[info['planeaxis']] = info['point']
            if test_point not in self:
                msg = f'Point {test_point} is outside the mesh.'
                raise ValueError(msg)

        # Determine the n tuple.
        if n is None:
            n = (self.n[info['axis1']], self.n[info['axis2']])

        # Build a mesh.
        p1s, p2s, ns = np.zeros(3), np.zeros(3), np.zeros(3)
        ilist = [info['axis1'], info['axis2'], info['planeaxis']]
        p1s[ilist] = (self.pmin[info['axis1']],
                      self.pmin[info['axis2']],
                      info['point'] - self.cell[info['planeaxis']]/2)
        p2s[ilist] = (self.pmax[info['axis1']],
                      self.pmax[info['axis2']],
                      info['point'] + self.cell[info['planeaxis']]/2)
        ns[ilist] = n + (1,)
        ns = dfu.array2tuple(ns.astype(int))

        plane_mesh = self.__class__(p1=p1s, p2=p2s, n=ns)
        plane_mesh.info = info  # Add info so it can be interpreted easier
        return plane_mesh

    def mpl(self, figsize=None):
        """Plots the mesh domain and the discretisation cell using a
        `matplotlib` 3D plot.

        Parameters
        ----------
        figsize : tuple, optional
            Length-2 tuple passed to the `matplotlib.pyplot.figure`
            function.

        Examples
        --------
        1. Visualising the mesh using `matplotlib`

        >>> import discretisedfield as df
        ...
        >>> p1 = (-6, -3, -3)
        >>> p2 = (6, 3, 3)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.mpl()

        """
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111, projection="3d")

        dfu.plot_box(ax, self.pmin, self.pmax, "b-", linewidth=1.5)
        dfu.plot_box(ax, self.pmin, np.add(self.pmin, self.cell),
                     "r--", linewidth=1)

        ax.set(xlabel=r"$x$", ylabel=r"$y$", zlabel=r"$z$")

    def k3d(self, colormap=dfu.colormap, plot=None, **kwargs):
        """Plots the mesh domain and emphasises the discretisation cell.

        The first element of `colormap` is the colour of the domain,
        whereas the second one is the colour of the discretisation
        cell. If `plot` is passed as a `k3d.plot.Plot`, plot is added
        to it. Otherwise, a new k3d plot is created. All arguments
        allowed in `k3d.voxels()` can be passed. This function is to
        be called in Jupyter notebook.

        Parameters
        ----------
        colormap : list, optional
            Length-2 list of colours in hexadecimal format. The first
            element is the colour of the domain, whereas the second
            one is the colour of the discretisation cell.
        plot : k3d.plot.Plot, optional
            If this argument is passed, plot is added to
            it. Otherwise, a new k3d plot is created.

        Examples
        --------
        1. Visualising the mesh using `k3d`

        >>> import discretisedfield as df
        ...
        >>> p1 = (-6, -3, -3)
        >>> p2 = (6, 3, 3)
        >>> cell = (1, 1, 1)
        >>> mesh = df.Mesh(p1=p1, p2=p2, cell=cell)
        >>> mesh.k3d()
        Plot(...)

        """
        plot_array = np.ones(tuple(reversed(self.n)))
        plot_array[0, 0, -1] = 2  # mark the discretisation cell
        dfu.voxels(plot_array, pmin=self.pmin, pmax=self.pmax,
                   colormap=colormap, plot=plot, **kwargs)

    def k3d_points(self, point_size=0.5, color=dfu.colormap[0],
                   plot=None, **kwargs):
        """Plots the points at discretisation cell centres.

        The size of points can be defined with `point_size` argument
        and their colours with `color`. If `plot` is passed as a
        `k3d.plot.Plot`, plot is added to it. Otherwise, a new k3d
        plot is created. All arguments allowed in `k3d.points()` can
        be passed. This function is to be called in Jupyter notebook.

        Parameters
        ----------
        point_size : float, optional
            The size of a single point.
        color : hex, optional
            Colour of a single point.
        plot : k3d.plot.Plot, optional
            If this argument is passed, plot is added to
            it. Otherwise, a new k3d plot is created.

        Examples
        --------
        1. Plotting discretisation cell centres using `k3d.points`

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (20, 20, 10)
        >>> n = (10, 10, 5)
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n)
        >>> mesh.k3d_points()
        Plot(...)

        """
        plot_array = np.array(list(self.coordinates))
        dfu.points(plot_array, point_size=point_size, color=color,
                   plot=plot, **kwargs)

    def k3d_regions(self, colormap=dfu.colormap, plot=None, **kwargs):
        """Plots the mesh domain and emphasises defined regions.

        The order of colours in `colormap` should be the same as the
        order of regions defined in
        `discretisedfield.Mesh.regions`. By default 6 colours are
        defined and for all additional regions, the colours will be
        generated randomly. If `plot` is passed as a `k3d.plot.Plot`,
        plot is added to it. Otherwise, a new k3d plot is created. All
        arguments allowed in `k3d.voxels()` can be passed. This
        function is to be called in Jupyter notebook.

        Parameters
        ----------
        colormap : list, optional
            List of colours in hexadecimal format. The order of
            colours should be the same as the order of regions defined
            in `discretisedfield.Mesh.regions`. By default 6 colours
            are defined.
        plot : k3d.plot.Plot, optional
            If this argument is passed, plot is added to
            it. Otherwise, a new k3d plot is created.

        Examples
        --------
        1. Visualising defined regions in the mesh

        >>> import discretisedfield as df
        ...
        >>> p1 = (0, 0, 0)
        >>> p2 = (100, 100, 100)
        >>> n = (10, 10, 10)
        >>> regions = {'r1': df.Region(p1=(0, 0, 0), p2=(50, 100, 100)),
        ...            'r2': df.Region(p1=(50, 0, 0), p2=(100, 100, 100))}
        >>> mesh = df.Mesh(p1=p1, p2=p2, n=n, regions=regions)
        >>> mesh.k3d_regions()
        Plot(...)

        """
        # Generate random colours if necessary
        if len(self.regions) > 6:
            for i in range(len(self.regions)-6):
                found = False
                while not found:
                    color = random.randint(0, 16777215)
                    found = True
                colormap.append(color)

        plot_array = np.zeros(self.n)
        for i, name in enumerate(self.regions.keys()):
            for index in self.indices:
                if self.index2point(index) in self.regions[name]:
                    plot_array[index] = i+1  # i+1 to avoid 0 value
        plot_array = np.swapaxes(plot_array, 0, 2)  # swap axes for k3d.voxels
        dfu.voxels(plot_array, pmin=self.pmin, pmax=self.pmax,
                   colormap=colormap, plot=plot, **kwargs)

    @property
    def _script(self):
        """This abstract method should be implemented by a specific
        calculator.

        Raises
        ------
        NotImplementedError
            If not implemented by a specific calculator.

        """
        raise NotImplementedError



import pytest
import numpy as np
import discretisedfield as df


def check_region(region):
    assert isinstance(region.pmin, tuple)
    assert len(region.pmin) == 3
    assert isinstance(region.pmax, tuple)
    assert len(region.pmax) == 3


class TestRegion:
    def setup(self):
        self.valid_args = [[(0, 0, 0), (5, 5, 5)],
                           [(-1, 0, -3), (5, 7, 5)],
                           [(0, 0, 0), (5e-9, 5e-9, 5e-9)],
                           [(-1.5e-9, -5e-9, 0), (1.5e-9, -15e-9, -10e-9)],
                           [(-1.5e-9, -5e-9, -5e-9), np.array((0, 0, 0))],
                           [[0, 5e-6, 0], (-1.5e-6, -5e-6, -5e-6)],
                           [(0, 125e-9, 0), (500e-9, 0, -3e-9)]]

        self.invalid_args = [[('1', 0, 0), (1, 1, 1)],
                             [(-1.5e-9, -5e-9, 'a'), (1.5e-9, 15e-9, 16e-9)],
                             [(-1.5e-9, -5e-9, 0), (1.5e-9, 16e-9)],
                             [(-1.5e-9, -5e-9, 0), (1.5e-9, 15e-9, 1+2j)],
                             ['string', (5, 1, 1e-9)]]

    def test_init(self):
        p1 = (0, -4, 16.5)
        p2 = (15, -6, 11)
        region = df.Region(p1=p1, p2=p2)

        check_region(region)
        assert region.pmin == (0, -6, 11)
        assert region.pmax == (15, -4, 16.5)

    def test_init_valid_args(self):
        for p1, p2 in self.valid_args:
            region = df.Region(p1=p1, p2=p2)

            check_region(region)

    def test_init_invalid_args(self):
        for p1, p2 in self.invalid_args:
            with pytest.raises((TypeError, ValueError)):
                # Exceptions are raised by descriptors.
                region = df.Region(p1=p1, p2=p2)

    def test_zero_domain_edge_length(self):
        args = [[(0, 100e-9, 1e-9), (150e-9, 100e-9, 6e-9)],
                [(0, 101e-9, -1), (150e-9, 101e-9, 0)],
                [(10e9, 10e3, 0), (0.01e12, 11e3, 5)]]

        for p1, p2 in args:
            with pytest.raises(ValueError) as excinfo:
                region = df.Region(p1=p1, p2=p2)
            assert 'is zero' in str(excinfo.value)

    def test_repr(self):
        p1 = (-1, -4, 11)
        p2 = (15, 10.1, 12.5)
        region = df.Region(p1=p1, p2=p2)

        rstr = 'Region(p1=(-1.0, -4.0, 11.0), p2=(15.0, 10.1, 12.5))'
        assert repr(region) == rstr

    def test_contains(self):
        p1 = (0, 10e-9, 0)
        p2 = (10e-9, 0, 20e-9)
        region = df.Region(p1=p1, p2=p2)

        assert (0, 0, 0) in region
        assert (10e-9, 10e-9, 10e-9) in region
        assert (5e-9, 5e-9, 5e-9) in region
        assert (11e-9, 11e-9, 11e-9) not in region
        assert (-1e-9, -1e-9, -1e-9) not in region
